/* Copyright (c) 2003-2006 ActiveState Software Inc.
   See the file LICENSE.txt for licensing information. */

/* Dialog to display when license is invalid
 * Basically this works just like the customButtons dialog but has to
 * special layout.
 *
 * Usage:
 *  All dialog interaction is done via an object passed in and out as the first
 *  window argument: window.arguments[0]. All these arguments are optional.
 *      .message        the message to show.
 *      .buttons        list of strings naming buttons to use
 *      .infos          list of strings identifying what information blocks
 *                      to show; valid values are "beta", "purchase",
 *                      "winnt", "aspn"
 *      .response       the default response, must be one of the strings in
 *                      "buttons"
 *      .title          the dialog title
 *  On return window.arguments[0] has:
 *      .response       the name of button used
 */

var log = getLoggingMgr().getLogger("dialogs.invalidLicense");
//log.setLevel(LOG_DEBUG);

var gAcceptButton = null;
var gExtra1Button = null;
var gExtra2Button = null;
var gCancelButton = null;

var gInfoSvc = Components.classes["@activestate.com/koInfoService;1"].
      getService(Components.interfaces.koIInfoService);
var gPrefSvc = Components.classes["@activestate.com/koPrefService;1"].
                getService(Components.interfaces.koIPrefService);
var gPrefs = gPrefSvc.prefs; //XXX Don't think this is used.


//---- interface routines for XUL
// Check the current license validity and interact with the user accordingly.
//
//  "_mode" is an internal variable that should be used only for testing.
//
// This returns true iff the user's Komodo license is valid. Note that this
// method may involve installing a valid license as a side-effect.
//
function licenseCheckOnOpen(_mode /* =null */)
{
    if (typeof(_mode) == "undefined") _mode = null;
    var err;
    if (_mode || !gInfoSvc.licenseValid) {
        // Determine what "mode" we are in.
        var mode;
        if (_mode) {
            mode = _mode;
        } else {
            // - A build which includes a siloed non-expiring license.
            if (gInfoSvc.siloedLicense && !gInfoSvc.siloedLicenseExpires) {
                mode = "siloed non-expiring license";
            }
            // - A build which includes a siloed expiring license (presumably
            //   this is a beta) and the period has expired.
            else if (gInfoSvc.siloedLicense && gInfoSvc.siloedLicenseExpires
                     && gInfoSvc.siloedLicenseExpired)
            {
                mode = "siloed expired license";
            }
            // - A build which includes a siloed expiring license (presumably
            //   this is a beta) and the period has NOT expired.
            else if (gInfoSvc.siloedLicense && gInfoSvc.siloedLicenseExpires
                     && !gInfoSvc.siloedLicenseExpired)
            {
                mode = "siloed expiring license";
            }
            // - A non-final build withOUT a siloed license.
            else if (gInfoSvc.version.indexOf('-') != -1) {
                mode = "non-final build without license";
            }
            // - A final build withOUT a siloed license.
            else {
                mode = "final build without license";
            }
        }

        // Determine the appropriate UI for the current mode.
        // - Setting "choice" in the code block implies that no dialog
        //   should be shown, the given choice should just be assumed.
        var msg=null, buttons=[], infos=[], title=null, choice=null;
        switch (mode) {
        case "siloed non-expiring license":
        case "siloed expiring license":
            // in non error conditions, we should never get siloed modes
            // but we'll pop up an nice dialog just in case
        case "siloed expired license":
            var verInfo = "Komodo " + gInfoSvc.productType + " " +
                          gInfoSvc.version;
            msg = "The license period for this Komodo installation "+
                  "("+verInfo+") has expired.";
            if (gInfoSvc.version.indexOf('-') != -1) {
                buttons = ["Download New Komodo Beta (if available)",
                           "Download Komodo Trial",
                           "Purchase Komodo",
                           "Close"];
            } else {
                buttons = ["Purchase Komodo",
                           "Download Komodo Trial",
                           "Renew ASPN Membership",
                           "Close"];
            }
            infos = ["purchase", "aspn"];
            title = "Expired Komodo Installation";
            break;
        case "non-final build without license":
            msg = "You do not currently have a valid license installed for "+
                  "this Komodo version.";
            buttons = ["Get Beta License (if available)",
                       "Download Komodo Trial",
                       "Purchase Komodo",
                       "Close"];
            infos = ["purchase",
// #if PLATFORM == "win"
                     "winnt",
// #endif
                     "beta"];
            break;
        case "final build without license":
            msg = "You do not currently have a valid license installed for "+
                  "this Komodo version.";
            buttons = ["Download Komodo Trial",
                       "Purchase Komodo",
                       "Renew ASPN Membership",
                       "Close"];
            infos = ["purchase",
// #if PLATFORM == "win"
                     "winnt",
// #endif
                     "aspn"];
            break;
        default:
            err = "unexpected invalid license mode: '"+mode+"'\n";
            dialog_internalError(err, err);
            return null;
        }
    }
    var obj = new Object();
    obj.message = msg;
    obj.buttons = buttons;
    obj.infos = infos;
    obj.title = title;
    return obj;
}

function licenseCheckOnClose(choice) {
    // Handle the user's choice.
    switch(choice) {
    case "Get Beta License (if available)":
    case "Download New Komodo Beta (if available)":
        browse_OpenUrlInDefaultBrowser("http://www.ActiveState.com/Komodo/Beta");
        break;
    case "Purchase Komodo":
        browse_OpenUrlInDefaultBrowser("http://www.ActiveState.com/Komodo/Purchase");
        break;
    case "Download Komodo Trial":
        browse_OpenUrlInDefaultBrowser("http://www.ActiveState.com/Komodo/Trial");
        break;
    case "Renew ASPN Membership":
        browse_OpenUrlInDefaultBrowser("http://www.ActiveState.com/renewal_info");
        break;
    case "Close":
        // do nothing
        break;
    default:
        err = "unexpected invalid license choice: '"+choice+"'\n";
        dialog_internalError(err, err);
        return false;
    }

    return false;
}

function OnLoad()
{
    try {
        trace_onload();
        var i;
        var dialog = document.getElementById("dialog-invalidlicense")
        gAcceptButton = dialog.getButton("accept");
        gExtra1Button = dialog.getButton("extra1");
        gExtra2Button = dialog.getButton("extra2");
        gCancelButton = dialog.getButton("cancel");
        gCancelButton.setAttribute("label", "Close");
        gCancelButton.setAttribute("accesskey", "C");
        var buttonWidgets = [gAcceptButton, gExtra1Button, gExtra2Button];

        // .message
        var messageWidget = document.getElementById("message");
        
        var config = licenseCheckOnOpen();
        if (!config) {
            window.close();
            return;
        }
        var message = config.message;
        if (typeof message != "undefined" && message != null) {
            var textNode = document.createTextNode(message);
            messageWidget.appendChild(textNode);
        } else {
            messageWidget.setAttribute("collapsed", "true");
        }

        // .buttons
        var buttons = config.buttons;
        if (typeof buttons == "undefined" || buttons == null) {
            var msg = "Internal Error: illegal 'buttons' value: '"+buttons+"'.";
            log.error(msg);
            alert(msg);
            window.close();
        } else if (buttons.length == 0 || buttons.length > 4) {
            var msg = "Internal Error: illegal number of buttons: '"+buttons.length+"'.";
            log.error(msg);
            alert(msg);
            window.close();
        } else if (buttons.length == 4 && buttons[3] != "Close") {
            var msg = "Internal Error: there are 4 buttons to display but "
                      +"the last one is not Close.";
            log.error(msg);
            alert(msg);
            window.close();
        } else if (buttons[buttons.length-1] == "Close") {
            for (i = 0; i < buttons.length-1; ++i) {
                buttonWidgets[i].setAttribute("label", buttons[i]);
            }
            for (i = buttons.length-1; i < buttonWidgets.length; ++i) {
                buttonWidgets[i].setAttribute("collapsed", true);
            }
        } else {
            for (i = 0; i < buttons.length; ++i) {
                buttonWidgets[i].setAttribute("label", buttons[i]);
            }
            for (i = buttons.length; i < buttonWidgets.length; ++i) {
                buttonWidgets[i].setAttribute("collapsed", true);
            }
            gCancelButton.setAttribute("collapsed", true);
        }

        // .infos
        var widget;
        var infos = config.infos;
        if (typeof infos == "undefined" || infos == null) infos = [];
        for (i = 0; i < infos.length; ++i) {
            widget = document.getElementById(infos[i]+"-info");
            if (widget && widget.hasAttribute("collapsed")) {
                widget.removeAttribute("collapsed");
            }
        }

        // .response
        /*
        var response = window.arguments[0].response;
        if (typeof response == "undefined" || response == null) {
            // gAcceptButton (the first one) is already the default
        } else {
            gAcceptButton.removeAttribute("default");
            for (i = 0; i < buttonWidgets.length; ++i) {
                if (buttonWidgets[i].getAttribute("label") == response) {
                    buttonWidgets[i].setAttribute("default", true);
                    buttonWidgets[i].focus();
                }
            }
            // Note: we are not raising an error or warning if the default response
            // does not match one of the given button names. No big deal.
        }
        */
        // .title
        if (typeof config.title != "undefined" &&
            config.title != null) {
            document.title = config.title;
        } else {
            document.title = "Invalid Komodo License";
        }

        window.sizeToContent();
        // Cannot use dialog.moveToAlertPosition() because that requires
        // a loaded opener window, which this dialog typically does not have.
        dialog.centerWindowOnScreen();
    } catch(ex) {
        log.error(ex, "Error loading invalid license dialog.");
    }
    window.getAttention();
}


function Accept()
{
    try {
        licenseCheckOnClose(gAcceptButton.getAttribute("label"));
    } catch(ex) {
        log.error(ex);
    }
    return true;
}

function Extra1()
{
    try {
        // This is one of the "extra" <dialog/> buttons. Only the "accept" and
        // "cancel" button actions will automatically close the window, so we have
        // to do it manually here.
        licenseCheckOnClose(gExtra1Button.getAttribute("label"));
        window.close();
    } catch(ex) {
        log.error(ex);
    }
    return true;
}

function Extra2()
{
    try {
        // This is one of the "extra" <dialog/> buttons. Only the "accept" and
        // "cancel" button actions will automatically close the window, so we have
        // to do it manually here.
        licenseCheckOnClose(gExtra2Button.getAttribute("label"));
        window.close();
    } catch(ex) {
        log.error(ex);
    }
    return true;
}

function Cancel()
{
    try {
        licenseCheckOnClose(gCancelButton.getAttribute("label"));
    } catch(ex) {
        log.error(ex);
    }
    return true;
}

