/* Copyright (c) 2006 ActiveState Software Inc.
   See the file LICENSE.txt for licensing information. */

/* Handling of VI editor commands for Komodo.
 *
 * See:
 *  http://www.vim.org
 *
 * Author:
 *   Todd Whiteman
 *
 * Overview of how it works:
 *  - The vi-controller loads as part of the keybinding load / initialization.
 *  - It is enabled by having one or more cmd_vim_* commands being defined in
 *    the current keybinding scheme.
 *  - It intercepts keypress events within scintilla (when enabled), and may
 *    capture and swallow keypress events there.
 *  - The keybinding manager handles the mapping of vi key codes to the actual
 *    commands.
 *  - May override existing non-vi commands, catching the commands as they are
 *    called and can result in a different action if needed.
 *  - Remembers last commands, command mode history, search patterns etc...
 *
 * Commands:
 *  - Keybinds are setup in the "schemes/_vi.(unprocessed.)kkf file"
 *  - All commands go through the vim_doCommand() handler. This is so the
 *    vi controller knows what commands have been performed and what data was
 *    used for the commands. Mainly used for repeat, redo and undo operations.
 *  - Most of the vi commands map to existing komodo commands, so a 1-to-1
 *    mapping from vim_doCommand() to command_doCommand() exists. The ones that
 *    don't have a mapping are marked as VimController.SPECIAL_COMMAND, and are
 *    then implemented as JS functions (same name to the cmd) in this file.
 *
 * Mode:
 *  - Vi has a number of states/modes (search, command, repeat) that need to
 *    remember their state information. This is handled using one class object,
 *    called the VimController (gVimController is the global variable).
 *
 */
//
//---- BUGS:
//  - various bugs when working at beginning or end of lines, as vim treats the
//    end of line a little specially.
//  - repeat count of the form [n]x[m]y, where both m and n are supplied will
//    not work as expected (n * m xy), but will interpret as (nm xy).
//
//---- PERFORMANCE:
//  - This code is another layer on top of command_doCommand code, so it has
//    a large overhead for every command performed. Some multi-commands may
//    need to be moved into the scintilla python wrapper as one-command.
//
//---- TODO:
//  - support for more command mode ":" strings
//  - read in .vimrc to add certain customized workings
//  - implement registers??
//  - implement markers??
//
//  See http://www.lagmonster.org/docs/vi.html for simple overview of most vi
//  features
//

var vimlog = getLoggingMgr().getLogger('vimlog');
//vimlog.setLevel(LOG_DEBUG);

var gVimController;

function VimController_onload()
{
    vimlog.info("VimController_onload");
    try {
        gVimController = new VimController();
    } catch (e) {
        vimlog.error(e);
    }
}

function VimController() {
    try {
        this.enabled = false;
        this._mode = VimController.MODE_NORMAL;
        this._operationFlags = VimController.OPERATION_NONE; // Special operation type
        this._copyMode = VimController.COPY_NORMAL; // Type of copy operation done
        this._visualMode = VimController.VISUAL_CHAR;  // Visual selection mode
        this._internalBuffer = "";          // Internal clipboard
        // Anchor and currentPos are used to store command movements
        this._currentPos = 0;
        this._anchor = 0;
        // Undo/redo/repeat
        this._repeatCount = 0;              // Number of times to repeat the command
        this._lastMode = VimController.MODE_NORMAL;
        this._lastOperationFlags = VimController.OPERATION_NONE; // Last operation type
        this._lastModifyCommand = "";       // Last modify command performed
        this._lastRepeatCount = 1;          // Number of times command was repeated
        this._lastInsertedKeycodes = [];    // Array of chars used in last command
        // Searching
        this._findSvc = Components.classes["@activestate.com/koFindService;1"].
                               getService(Components.interfaces.koIFindService);

        this._inputBuffer = null;
        this._inputBuffer_active = false;
        this._inputBuffer_history = [];        // History of previous values
        this._inputBuffer_historyPosition = 0; // Where we are up to in history
        this._inputBuffer_savedText = "";      // Saved edited input buffer

        this._searchString = "";
        this._searchDirection = VimController.SEARCH_FORWARD;
        this._searchHistory = [];           // Array of past search strings
        this._findCharDirection = VimController.SEARCH_FORWARD;
        // Command history
        this._commandHistory = [];            // Array of past command strings
        //this._commandHistoryPosition = 0;   // Where we are up to in history, unsed for undo/redo
        //this._commandHistoryRepeats = [];   // Array of past repeat counts
        // Command history for command mode
        this._commandModeHistory = [];      // Array of past command mode strings
        // Initialize the command details service
        this._commandDetails = Components.classes['@activestate.com/koViCommandDetail;1'].
                                getService(Components.interfaces.koIViCommandDetail);
    } catch (e) {
        vimlog.exception(e);
    }
}

// The following two lines ensure proper inheritance (see Flanagan, p. 144).
VimController.prototype = new Controller();
VimController.prototype.constructor = VimController;

VimController.prototype.destructor = function() {
}

// Used to capture mouseup/selection movements when in visual mode
var gVim_onWindowMouseUpHandler = function(event) {
    // We only want events going to the view
    if ((event.target.nodeName == 'view') &&
        (!('originalTarget' in event) ||
         (event.originalTarget.nodeName != "html:input"))) {
        gVimController.updateModeIfBufferHasSelection(true);
    }
}

VimController.prototype.enable = function(enabled) {
    if (enabled && !this.enabled) {
        window.addEventListener('mouseup', gVim_onWindowMouseUpHandler, true);
        this.loadOverlay();
    } else if (!enabled && this.enabled) {
        window.removeEventListener('mouseup', gVim_onWindowMouseUpHandler, true);
        this.unloadOverlay();
    }
    this.updateViInternals();
}

VimController.prototype.NAME = "VimController";
VimController.MODE_NORMAL  = 0;
VimController.MODE_INSERT  = 1;
VimController.MODE_SELECT  = 2;
VimController.MODE_SEARCH  = 3;
VimController.MODE_COMMAND = 4; // ":" Has been used
VimController.MODE_VISUAL  = 5; // User is in visual selection mode
//VimController.MODE_DELETE_COMMAND = 6;  // Handle delete command
//VimController.MODE_CHANGE_COMMAND = 7;  // Handle special case of c9b cases?
VimController.MODE_REPLACE_CHAR = 8;  // Used for vi "r" command
VimController.MODE_OVERTYPE     = 9;  // Used for vi "R" command
VimController.MODE_FIND_CHAR    = 10; // Used for vi "s" command

// Matching string names for _mode
VimController.MODE_NAMES = [
    "Normal",
    "Insert",
    "Select",
    "Search",
    "Command",
    "Visual",
    "",
    "",
    "Replace one char",
    "Overtype",
    "Find one char",
    ""
];

// Operation flags used in NORMAL_MODE, mostly used in combination with a
// movement. Such as "d w" delete word, "y y" yank line (a special case)
VimController.OPERATION_NONE = 0;
VimController.OPERATION_YANK = 1;
VimController.OPERATION_DELETE = 2;
VimController.OPERATION_CHANGE = 3;  // Combination of yank and delete

VimController.OPERATION_NAMES = [
    "None",
    "Yank",
    "Change",
    "Delete",
    ""
];

// Search flags, used by _searchDirection and _findCharDirection
VimController.SEARCH_FORWARD  = 0;
VimController.SEARCH_BACKWARD = 1;

// Used by _visualMode
VimController.VISUAL_CHAR  = 0;     // Char grained selection
VimController.VISUAL_LINE  = 1;     // Line grained selction
VimController.VISUAL_BLOCK = 2;     // Block grained selection

VimController.VISUAL_NAMES = [
    "Char",
    "Line",
    "Block",
    ""
];

// Command flags
VimController.NO_REPEAT_ACTION  = 0x1;
VimController.REPEATABLE_ACTION = 0x2;    // Can be repeated
VimController.ENTER_MODE_INSERT = 0x4;
VimController.MODIFY_ACTION     = 0x8;    // Action modifies document
VimController.COPY_NORMAL       = 0x10;   // Normal copy / paste handling
VimController.COPY_LINES        = 0x20;   // Special line copy / paste handling
VimController.COPY_ACTION       = VimController.COPY_NORMAL |
                                  VimController.COPY_LINES;
VimController.SPECIAL_REPEAT_HANDLING = 0x40;   // Special handling of repeats
VimController.MOVEMENT_ACTION   = 0x80;   // If this command moves the cursor
VimController.WORKS_IN_VISUAL_MODE = 0x100; // Can be used in visual mode
VimController.CANCELS_VISUAL_MODE = 0x200; // Can be used in visual mode
VimController.CHOOSE_CARET_X =  0x400;    // Sets the caret X position after the command

// Used as representation that this command is specially handled
// Commands with this value must have a matching JS function of the same name.
VimController.SPECIAL_COMMAND = null;

// Load the vi overlay
VimController.prototype.loadOverlay = function(callback) {
    //dump("VimController.loadOverlay\n");
    window.controllers.appendController(this);
    if (!callback) {
        callback = {
            observe : function (subject, topic, data) {
                if (topic == 'xul-overlay-merged') {
                    //command_onload();
                    var vimset = gKeybindingMgr.document.getElementById('cmdset_vim');
                    //dump("vimset: " + vimset + "\n");
                    //command_updateCommandset(vimset);
                    // Force keybindings to reload it's command configuration
                    gKeybindingMgr.parseGlobalData();
                }
            }
        };
    }
    // Enable vi keybindings - load the vi command overlay
    //dump("Loading vi overlay\n");
    gKeybindingMgr.document.loadOverlay("chrome://komodo/content/keybindings/vi-commands-overlay.xul", callback);
    this.enabled = true;
}

// Remove the vi overlay commands
VimController.prototype.unloadOverlay = function() {
    //dump("VimController.unloadOverlay\n");
    // Disable vi keybindings - remove the vi commandset
    var vimset = gKeybindingMgr.document.getElementById('cmdset_vim');
    if (vimset && vimset.parentNode) {
        var parentNode = vimset.parentNode;
        //dump("Removing vimset: " + vimset + "\n");
        parentNode.removeChild(vimset);
        // Force keybindings to reload it's command configuration
        gKeybindingMgr.parseGlobalData();
    }
    window.controllers.removeController(this);
    this.enabled = false;
}

// Controller related code - handling of commands
VimController.prototype.isCommandEnabled = function(cmdName) {
    if (cmdName.substr(0, 8) == "cmd_vim_")
        return true;
    return false;
}

VimController.prototype.supportsCommand = VimController.prototype.isCommandEnabled;

VimController.prototype.doCommand = function(cmdName) {
    return vim_doCommand(cmdName);
}

VimController.prototype.onBufferFocus =
VimController.prototype.updateViInternals = function() {
    this.updateCaretStyle();
    this.updateModeIfBufferHasSelection(false);
}

VimController.prototype.updateStatusBarMessage = function() {
    // Add message to status bar informing of the mode change
    var msg = "Vi Mode: " + VimController.MODE_NAMES[this.mode];
    if (this.mode == VimController.MODE_VISUAL) {
        // Show which visual mode it is
        msg += " " + VimController.VISUAL_NAMES[this._visualMode];
    }
    window.StatusBar_AddMessage(msg, "vimode", 0, false);
}

VimController.prototype.updateCaretStyle = function() {
    if (typeof(gViewMgr) != 'undefined' && gViewMgr && gViewMgr.currentView &&
        gViewMgr.currentView.document && gViewMgr.currentView.scintilla) {
        var scimoz = gViewMgr.currentView.scintilla.scimoz;
        if ((this._mode == VimController.MODE_INSERT) ||
            (this._mode == VimController.MODE_OVERTYPE)) {
            scimoz.caretWidth = 2;
            //scimoz.caretStyle = scimoz.CARETSTYLE_LINE;
            //dump("VimController:: Making caret a LINE style\n");
        } else {
            scimoz.caretWidth = 1;
            //scimoz.caretStyle = scimoz.CARETSTYLE_BLOCK;
            //dump("VimController:: Making caret a BLOCK style\n");
        }
    }
}

VimController.prototype.updateModeIfBufferHasSelection = function(fromMouseUp) {
    if (typeof(gViewMgr) != 'undefined' && gViewMgr && gViewMgr.currentView &&
        gViewMgr.currentView.document && gViewMgr.currentView.scintilla) {
        var scimoz = gViewMgr.currentView.scintilla.scimoz;
        if ((this.mode != VimController.MODE_VISUAL) &&
            (scimoz.currentPos != scimoz.anchor)) {
            // Default will be char visual mode
            this.mode = VimController.MODE_VISUAL;
            this._visualMode = VimController.VISUAL_CHAR;
        // If we are in visual mode already, it might have been clicked away
        } else if (fromMouseUp && (this.mode == VimController.MODE_VISUAL) &&
                   (scimoz.currentPos == scimoz.anchor)) {
            this.mode = VimController.MODE_NORMAL;
        }
    }
}

// Special class getters and setters
VimController.prototype.__defineGetter__("mode", function() { return this._mode });
VimController.prototype.__defineSetter__("mode", function(new_mode) {
    var old_mode = this._mode;
    // Now update the mode
    this._mode = new_mode;

    if (old_mode != VimController.MODE_NORMAL) {
        // Reset the repeat count
        this.repeatCount = 0;
    }

    // If the mode changed, update the vi characteristics
    if (old_mode != new_mode) {
        vimlog.debug("Vi changed mode from " + old_mode + " to " + new_mode);
        // Remember the last mode
        this._lastMode = old_mode;
        // Set operation type back to none
        this.operationFlags = VimController.OPERATION_NONE;
        if (old_mode == VimController.MODE_OVERTYPE) {
            // Turning off overtype mode
            gViewMgr.currentView.scintilla.scimoz.overtype = 0;
        } else if ((new_mode == VimController.MODE_NORMAL) &&
                   (old_mode == VimController.MODE_VISUAL)) {
            // Remove any selection when leaving visual mode for normal mode
            var scimoz = gViewMgr.currentView.scintilla.scimoz;
            // scimoz.cancel() used to ensure removal of the selectionmode
            // http://bugs.activestate.com/show_bug.cgi?id=65580
            scimoz.cancel();
            scimoz.setSel(-1, scimoz.currentPos);  // anchorPos of -1 means remove any selection
        }
        this.updateCaretStyle();
        this.updateStatusBarMessage();
    }
});

VimController.prototype.__defineGetter__("repeatCount", function() { return this._repeatCount });
VimController.prototype.__defineSetter__("repeatCount", function(new_count) {
    if (this._repeatCount != new_count) {
        vimlog.debug("Setting repeatCount to: " + new_count);
        var msg = "Vi Mode: " + VimController.MODE_NAMES[this.mode];
        if (this.mode == VimController.MODE_VISUAL) {
            // Show which visual mode it is
            msg += " " + VimController.VISUAL_NAMES[this._visualMode];
        }
        if (new_count > 0) {
            window.StatusBar_AddMessage(msg + " (Repeat " + new_count + " times)",
                                        "vimode", 0, false);
        } else {
            window.StatusBar_AddMessage(msg, "vimode", 0, false);
        }
        this._repeatCount = new_count;
    }
});

VimController.prototype.__defineGetter__("operationFlags", function() { return this._operationFlags });
VimController.prototype.__defineSetter__("operationFlags", function(new_operationFlags) {
    if (this._operationFlags != new_operationFlags) {
        vimlog.debug("vi:: changing operationFlags from: "+
                     this.operationFlags+", to: "+new_operationFlags);
        this._operationFlags = new_operationFlags;
    }
});

VimController.prototype.__defineGetter__("inputBuffer",
    function() {
        if (!this._inputBuffer) {
            this._inputBuffer = document.getElementById("vim-input-buffer");
        }
        return this._inputBuffer;
    }
);


VimController.prototype.resetCommandSettings = function() {
    this.operationFlags = VimController.OPERATION_NONE;
}

VimController.prototype.updateVisualModeSelection = function(scimoz,
                                                             currentPos,
                                                             anchorPos) {
    vimlog.debug("updateVisualModeSelection:: currentPos: "+currentPos+
                 ", anchorPos: "+anchorPos);
    if (this._visualMode == VimController.VISUAL_LINE) {
        scimoz.selectionMode = scimoz.SC_SEL_LINES;
    } else if (this._visualMode == VimController.VISUAL_BLOCK) {
        scimoz.selectionMode = scimoz.SC_SEL_RECTANGLE;
    }
    if (typeof(currentPos) != 'undefined' && currentPos != null) {
        scimoz.currentPos = currentPos;
    }
    if (typeof(anchorPos) != 'undefined' && anchorPos != null) {
        scimoz.anchor = anchorPos;
    }
    vimlog.debug("updateVisualModeSelection:: new currentPos: "+currentPos+
                 ", new anchorPos: "+anchorPos);
}

VimController.prototype.updateCursorAndSelection = function(scimoz,
                                                            oldCurrentPos,
                                                            newCurrentPos,
                                                            anchorPos,
                                                            setCaretX) {
    // Update the current position with scimoz if it's changed by vi
    if (gVimController.mode == VimController.MODE_VISUAL) {
        // Save these values, so we can support visual mode
        gVimController.updateVisualModeSelection(scimoz,
                                                 newCurrentPos,
                                                 anchorPos);
    } else if ((newCurrentPos != null) && (newCurrentPos != oldCurrentPos)) {
        vimlog.debug("updateCursorAndSelection:: Setting currentPos: "+newCurrentPos+
                     ", oldCurrentPos: "+oldCurrentPos);
        scimoz.currentPos = newCurrentPos;
        // Set anchor to the same spot, so we don't have a selection
        vimlog.debug("updateCursorAndSelection:: NOT Visual Mode: Removing any selection");
        scimoz.anchor = newCurrentPos;
    } else {
        vimlog.debug("updateCursorAndSelection:: Nothing changed");
    }
    if (setCaretX) {
        scimoz.chooseCaretX();
    }
}

//
// Key press handling
//

// Handling keypress, called from the main keybinding manager.
// We return true when we don't want the key binding manager to handle this
// keypress (i.e. so it does not match to a keybinding).
// This is used to allow a vi repeat count when we are in a prefix capture
// mode, i.e to allow for "d5w" command syntax. We will get the event again in
// the main keypressHandler function below, when called by xbl views-editor.
VimController.prototype.inSpecialModeHandler = function(event) {
    try {
        // If there is a key_handler (like iSearch), we don't want the
        // vi keybingings handling the event.
        // http://bugs.activestate.com/show_bug.cgi?id=65406
        if (gViewMgr.currentView && gViewMgr.currentView.scintilla &&
            gViewMgr.currentView.scintilla.key_handler) {
            return true;
        }

        if ((this.mode == VimController.MODE_NORMAL) ||
            (this.mode == VimController.MODE_VISUAL)) {
            var charCode = event.charCode;
            var specialKeyUsed = event.ctrlKey || event.altKey || event.metaKey;
            if (!specialKeyUsed && 
                (((charCode == event.DOM_VK_0) &&
                  (this.repeatCount > 0)) ||
                 ((charCode >= event.DOM_VK_1) &&
                  (charCode <= event.DOM_VK_9)))) {
                return true;
            }
        }
    } catch (e) {
        vimlog.exception(e);
    }
    return false;
}


// Used to map commands into vi commands when the vi mode == MODE_NORMAL
var normalModeCommandMap = {
    'cmd_left':                 'cmd_vim_left',
    'cmd_right':                'cmd_vim_right',
    'cmd_linePrevious':         'cmd_vim_linePrevious',
    'cmd_lineNext':             'cmd_vim_lineNext',
    'cmd_pageDown':             'cmd_vim_pageDown',
    'cmd_pageUp':               'cmd_vim_pageUp',
    'cmd_lineScrollUp':         'cmd_vim_lineScrollUp',
    'cmd_lineScrollDown':       'cmd_vim_lineScrollDown',
    'cmd_newline':              'cmd_vim_lineNextHome',
    'cmd_backSmart':            'cmd_vim_left',
    'cmd_delete':               'cmd_vim_cutChar'
};

// Vi handles command
VimController.prototype.handleCommand = function(event, commandname, keylabel, multikey) {
    try {
        /* Ensure it's going to a vi controlled scintilla component */
        if (!this.enabled || event.target.nodeName != 'view') {
            return false;
        }

        if (commandname == 'cmd_cancel') {
            vim_doCommand('cmd_vim_cancel');
            return false;
        } else if ((this.mode == VimController.MODE_NORMAL) ||
                   (this.mode == VimController.MODE_VISUAL)) {
            if (commandname in normalModeCommandMap) {
                var mappedCommand = normalModeCommandMap[commandname];
                //dump("commandname in handledCommandMap: true" + "\n");
                vim_doCommand(mappedCommand);
                return true;
            }
        } else if (this.mode == VimController.MODE_INSERT ||
                   this.mode == VimController.MODE_OVERTYPE) {
            if (commandname == 'cmd_backSmart') {
                // Remove the last character from recorded chars
                //dump("Removing last inserted key\n");
                this._lastInsertedKeycodes.pop();
            } else if (commandname != 'cmd_cancel') {
                // Any other characters (left, right, etc...) reset the buffer
                //dump("Resetting keycodes back to empty\n");
                this._lastInsertedKeycodes = [];
            }
        }
    } catch (e) {
        vimlog.exception(e);
    }
    return false;
}


// Handling keypress, called from the views-editor xul widget.
// We return false when we want this key event to continue on to other elements.
// Note: Returning true means the event is stopped (handled).
VimController.prototype.handleKeypress = function(event) {
    try {
        var keyCode = event.keyCode;
        var charCode = event.charCode;
        var key_char = String.fromCharCode(charCode);
        var specialKeyUsed = event.ctrlKey || event.altKey || event.metaKey;

        if (key_char == "\0") {
            // This is just so the dump/print operation works
            key_char = "\\0";
        }
        //dump(this.NAME + "::sendKeypressToScintilla: mode: " + this.mode + ", charCode: '" + key_char + "' (" + charCode + "), keyCode: " + keyCode + "\n");
    
        if (this.mode == VimController.MODE_INSERT ||
            this.mode == VimController.MODE_OVERTYPE) {
            // Any normal character should be fine to send to scintilla as is
            // Record the input in case we need to repeat
            if (charCode != 0 && !specialKeyUsed) {
                this._lastInsertedKeycodes.push(key_char);
            }
            return false;

        } else if (this.mode == VimController.MODE_REPLACE_CHAR) {
            this.mode = VimController.MODE_NORMAL;
            if (charCode != 0) {
                // Only perform the deletion if there is a character to
                // replace it with.
                // XXX - Does not handle a count replacement (i.e. 5rx)
                command_doCommand('cmd_delete');
                this._lastInsertedKeycodes = [key_char];
                gVimController._lastModifyCommand = 'cmd_delete';
            }
            return false;

        } else if (this.mode == VimController.MODE_FIND_CHAR) {
            this.mode = this._lastMode;
            var charToFind = key_char;
            //dump("Wanting to find: " + charToFind + "\n");
            if (charToFind) {
                var scimoz = gViewMgr.currentView.scintilla.scimoz;
                var currentPos = scimoz.currentPos;
                var lineNo = scimoz.lineFromPosition(currentPos);
                if (this._findCharDirection == VimController.SEARCH_BACKWARD) {
                    var lineStartPos = scimoz.positionFromLine(lineNo);;
                    var text = scimoz.getTextRange(lineStartPos, currentPos);
                    var pos = text.lastIndexOf(charToFind);
                    if (pos >= 0) {
                        // Found it, lets move to it then
                        var newCurrentPos = lineStartPos + stringutils_sysUtils.byteLength(text.substr(0, pos));
                        vimlog.debug("FindChar:: currentPos: " + currentPos + ", pos: " + pos +", newCurrentPos: "+newCurrentPos);
                        this.updateCursorAndSelection(scimoz, null, newCurrentPos, null, true);
                    }
                } else {
                    var lineEndPos = scimoz.getLineEndPosition(lineNo);
                    var nextPos = scimoz.positionAfter(currentPos);
                    var text = scimoz.getTextRange(nextPos, lineEndPos);
                    var pos = text.indexOf(charToFind);
                    if (pos >= 0) {
                        // Found it, lets move to it then
                        var newCurrentPos = nextPos + stringutils_sysUtils.byteLength(text.substr(0, pos));
                        vimlog.debug("FindChar:: currentPos: " + currentPos + ", pos: " + pos +", newCurrentPos: "+newCurrentPos);
                        this.updateCursorAndSelection(scimoz, null, newCurrentPos, null, true);
                    }
                }
            }

        } else if ((this.mode == VimController.MODE_NORMAL) ||
                   (this.mode == VimController.MODE_VISUAL)) {
            // Special case for repeat counts, the repeat count gets set
            // when checking if the event should go to scintilla.
            // TAB is handled specially by the scintilla overlay.
            if (keyCode == event.DOM_VK_TAB) {
                if (this.mode == VimController.MODE_NORMAL) {
                    // Leave TAB keypress alone, let the scintilla-overlay handle it
                    return false;
                //} else { // Swallow it, it doesn't do anything in visual mode
                }
            } else if (!specialKeyUsed &&
                       (((charCode == event.DOM_VK_0) &&
                         (this.repeatCount > 0)) ||
                        ((charCode >= event.DOM_VK_1) &&
                         (charCode <= event.DOM_VK_9)))) {
                // Repeat counts, swallows this event
                this.repeatCount = (this.repeatCount * 10) + (charCode - event.DOM_VK_0);
            } else if (event.keyCode == 0 && !specialKeyUsed) {
                // Stop normal characters going to scintilla
                //dump("    stopping the event.\n");
                event.stopPropagation();
                event.preventDefault();
                event.cancelBubble = true;
                gKeybindingMgr.keypressHandler(event);
            } else {
                // Else, leave everything else alone
                return false;
            }

        } else if ((this.mode == VimController.MODE_SEARCH) ||
                   (this.mode == VimController.MODE_COMMAND)) {
            this.inputBufferFocus();
        }
    } catch (e) {
        vimlog.exception(e);
    }
    return true;
}

VimController.prototype.inputBufferStart = function (initialText)
{
    try {
        this._inputBuffer_active = true;
        if (initialText == ":") {
            this._inputBuffer_history = this._commandHistory;
        } else {
            this._inputBuffer_history = this._searchHistory;
        }
        this._inputBuffer_historyPosition = this._inputBuffer_history.length;
        this.inputBuffer.value = initialText;
        this.inputBuffer.setAttribute("collapsed", "false");
        this.inputBuffer.addEventListener('keypress', vim_InputBuffer_KeyPress, false);
        this.inputBuffer.focus();
    } catch (e) {
        vimlog.exception(e);
    }
}

VimController.prototype.inputBufferFocus = function ()
{
    try {
        if (this._inputBuffer_active) {
            this.inputBuffer.focus();
        }
    } catch (e) {
        vimlog.exception(e);
    }
}

VimController.prototype.inputBufferFinish = function ()
{
    try {
        // Return the contents of the input buffer and stop buffering.
        var contents = this.inputBuffer.value;
        this.inputBuffer.value = "";
        this.inputBuffer.setAttribute("collapsed", "true");
        this.inputBuffer.removeEventListener('keypress', vim_InputBuffer_KeyPress, false);
        this._inputBuffer_active = false;
        return contents;
    } catch (e) {
        vimlog.exception(e);
    }
    return "";
}

VimController.prototype.inputBufferPrevious = function ()
{
    try {
        if (this._inputBuffer_historyPosition == 0) {
            // No previous commands
            return;
        }
        if (this._inputBuffer_historyPosition == this._inputBuffer_history.length) {
            // We are at the end, save the current input buffer
            this._inputBuffer_savedText = this.inputBuffer.value;
        }
        this._inputBuffer_historyPosition -= 1;
        this.inputBuffer.value = this.inputBuffer.value[0] +
                   this._inputBuffer_history[this._inputBuffer_historyPosition];
    } catch (e) {
        vimlog.exception(e);
    }
}

VimController.prototype.inputBufferNext = function ()
{
    try {
        if (this._inputBuffer_historyPosition >= this._inputBuffer_history.length) {
            // We are at the end already
            return;
        }
        this._inputBuffer_historyPosition += 1;
        if (this._inputBuffer_historyPosition == this._inputBuffer_history.length) {
            // We are at the end, restore the saved input buffer
            this.inputBuffer.value = this._inputBuffer_savedText;
        } else {
            this.inputBuffer.value = this.inputBuffer.value[0] +
                       this._inputBuffer_history[this._inputBuffer_historyPosition];
        }
    } catch (e) {
        vimlog.exception(e);
    }
}

VimController.prototype.inputBufferSaveHistory = function (value)
{
    try {
        if (this._inputBuffer_historyPosition > 0) {
            // We have previous commands, make sure it's not the same as the
            // previous command.
            if (value == this._inputBuffer_history[this._inputBuffer_historyPosition])
                return;
        }
        this._inputBuffer_history.push(value);
    } catch (e) {
        vimlog.exception(e);
    }
}

//
// Handling copy and paste (now handled internally, this._internalBuffer)
//
//   scimoz       - Scintilla xpcom object
//   start        - start position in buffer
//   end          - end position in buffer
//   deleteRange  - remove this range once copied (true or false)
VimController.prototype.copyInternal = function (scimoz, start, end,
                                                 deleteRange /* false */) {
    try {
        if (start < end) {
            gVimController._internalBuffer += scimoz.getTextRange(start, end);
            if (deleteRange) {
                scimoz.targetStart = start;
                scimoz.targetEnd = end;
                scimoz.replaceTarget(0, "");
            }
        }
    } catch (e) {
        vimlog.exception(e);
    }
}

//VimController.prototype.saveClipboard = function () {
//    try {
//        this._systemClipboard = null;
//        var clip = Components.classes["@mozilla.org/widget/clipboard;1"].getService(Components.interfaces.nsIClipboard);
//        var trans = Components.classes["@mozilla.org/widget/transferable;1"].createInstance(Components.interfaces.nsITransferable);
//        if (!clip || !trans) return;
//        //trans.addDataFlavor("text/unicode");
//        clip.getData(trans, clip.kGlobalClipboard);
//        this._systemClipboard = trans;       // Saving system clipboard
//    } catch (e) {
//        vimlog.exception(e);
//    }
//}
//
//VimController.prototype.restoreClipboard = function () {
//    try {
//        var clip = Components.classes["@mozilla.org/widget/clipboard;1"].getService(Components.interfaces.nsIClipboard);
//        var trans = Components.classes["@mozilla.org/widget/transferable;1"].createInstance(Components.interfaces.nsITransferable);
//        if (!clip || !trans) return;
//        trans.addDataFlavor("text/unicode");
//        var data = Components.classes["@mozilla.org/supports-string;1"].createInstance(Components.interfaces.nsISupportsString);
//        if (data) {
//            data.data = text;
//            transferable.setTransferData("text/unicode", data, text.length * 2);
//            // Put on clipboard.
//            clipboard.setData(transferable, null,
//                              Components.interfaces.nsIClipboard.kGlobalClipboard);
//            clip.getData(trans, clip.kGlobalClipboard);
//            this._internalBuffer = trans;       // Saving system clipboard
//        }
//        if (this._systemClipboard) {
//            clip.setData(this._systemClipboard, clip.kGlobalClipboard);
//            this._systemClipboard = null;
//        }
//    } catch (e) {
//        vimlog.exception(e);
//    }
//}

//
// Handling searches
//
VimController.prototype.initSearchForward = function () {
    try {
        this.inputBufferStart("/");
        //this._searchString = "";
        this._searchDirection = VimController.SEARCH_FORWARD;
        this.mode = VimController.MODE_SEARCH;
    } catch (e) {
        vimlog.exception(e);
    }
}

VimController.prototype.initSearchBackward = function () {
    try {
        this.inputBufferStart("?");
        //this._searchString = "";
        this._searchDirection = VimController.SEARCH_BACKWARD;
        this.mode = VimController.MODE_SEARCH;
    } catch (e) {
        vimlog.exception(e);
    }
}

VimController.prototype.performSearch = function (scimoz, searchString, reverseDirection) {
    try {
        // Vi starts searching from the position after the cursor/currentPos, so
        // remember currentPos then move past the current position.
        var orig_currentPos = scimoz.currentPos;
        var searchDirection = this._searchDirection;
        if (reverseDirection) {
            searchDirection = Number(!searchDirection);
        }
        if (searchDirection == VimController.SEARCH_FORWARD) {
            scimoz.currentPos = scimoz.positionAfter(orig_currentPos);
        } else {
            scimoz.currentPos = scimoz.positionBefore(orig_currentPos);
        }

        if (!searchString) {
            // Search again using the last know search text
            searchString = MRU_get("find-patternMru");
        }
        //dump("searchString: '"+this._searchString+"'\n");
        if (!searchString) {
            window.StatusBar_AddMessage("Search: No search string has been set",
                                        "visearch", 3000, true);
            // Didn't find anything, ensure we move back to the start position
            scimoz.currentPos = orig_currentPos;
            return;
        }
        this._searchString = searchString;

        this._findSvc.options.searchBackward = (searchDirection == VimController.SEARCH_BACKWARD);
        this._findSvc.options.matchWord = false;
        this._findSvc.options.patternType = this._findSvc.options.FOT_SIMPLE;
        this._findSvc.options.caseSensitivity = this._findSvc.options.FOC_INSENSITIVE;
        var incrementalSearchContext = Components.classes["@activestate.com/koFindContext;1"]
                                        .createInstance(Components.interfaces.koIFindContext);
        incrementalSearchContext.type = this._findSvc.options.FCT_CURRENT_DOC;
        var findres = Find_FindNext(window, incrementalSearchContext,
                                    this._searchString,
                                    null,  // mode
                                    true,  // quiet
                                    true); // add pattern to find MRU
        if (findres == false) {
            window.StatusBar_AddMessage("Search: No occurences of '" +
                                        this._searchString +"' found",
                                        "visearch", 3000, true);
            // Didn't find anything, ensure we move back to the start position
            scimoz.currentPos = orig_currentPos;
        } else {
            // The search engine highlights the selection, we don't want this:
            //   http://bugs.activestate.com/show_bug.cgi?id=65586
            var new_currentPos = scimoz.anchor;
            scimoz.currentPos = new_currentPos;
            var msg = "Search: '" + this._searchString + "'";
            if ((searchDirection == VimController.SEARCH_FORWARD) &&
                (new_currentPos < orig_currentPos)) {
                msg += " (search hit BOTTOM, continuing at TOP)";
            } else if ((this._searchDirection == VimController.SEARCH_BACKWARD) &&
                       (new_currentPos > orig_currentPos)) {
                msg += " (search hit TOP, continuing at BOTTOM)";
            }
            window.StatusBar_AddMessage(msg, "visearch", 3000, false);
        }
    } catch (e) {
        vimlog.exception(e);
    }
}

VimController.prototype.performReplace = function (searchString, replaceString,
                                                   startLine, endLine,
                                                   options) {
    try {
        //dump("Vi performReplace: " + startPos + "," + endPos + "s/" +
        //     searchString + "/" + replaceString + "/" + options + "\n";
        // Set the context
        findContext = Components.classes[
            "@activestate.com/koRangeFindContext;1"]
            .createInstance(Components.interfaces.koIRangeFindContext);
        findContext.type = Components.interfaces.koIFindContext.FCT_SELECTION;
        // Work out the real scintilla buffer positions
        var scimoz = gViewMgr.currentView.scintilla.scimoz;
        findContext.startIndex = scimoz.positionFromLine(startLine);;
        findContext.endIndex = scimoz.getLineEndPosition(endLine);

        this._findSvc.options.searchBackward = false;
        this._findSvc.options.matchWord = false;
        this._findSvc.options.patternType = this._findSvc.options.FOT_SIMPLE;
        if (options.search('i') >= 0) {
            this._findSvc.options.caseSensitivity = this._findSvc.options.FOC_INSENSITIVE;
        } else {
            this._findSvc.options.caseSensitivity = this._findSvc.options.FOC_SENSITIVE;
        }

        if (options.search('g') >= 0) {
            // Global replace
            var findres = Find_FindNext(window,
                                        findContext,
                                        searchString,
                                        null,  // mode, defaults to find
                                        true,  // quite
                                        false); // add pattern to find MRU
            if (!findres) {
                msg = "'"+searchString+"' was not found in the specified range. No changes were made.";
                window.StatusBar_AddMessage(msg, "find", 3000, true);
            } else {
                Find_ReplaceAll(window,           /* editor */
                                findContext,      /* context */
                                searchString,     /* pattern */
                                replaceString,    /* replacement */
                                false);           /* showReplaceResults */
            }
        } else {
            // This one is a little trickier, for two reasons:
            // 1. Need to call Find_FindNext() before Find_Replace (thats how
            //    the find stuff works.
            // 2. Vi has to do this once for every line in the given range
            var numReplacements = 0;
            for (var lineNum = startLine; lineNum <= endLine; lineNum++) {
                // Update the real scintilla buffer positions for current line
                findContext.startIndex = scimoz.positionFromLine(lineNum);
                findContext.endIndex = scimoz.getLineEndPosition(lineNum);
                var findres = Find_FindNext(window,
                                            findContext,
                                            searchString,
                                            null,  // mode, defaults to find
                                            true,  // quite
                                            false); // add pattern to find MRU
                if (findres) {
                    // Using _ReplaceLastFindResult() to get around having to
                    // call Find_Replace(), which gives feedback for every
                    // call made
                    _ReplaceLastFindResult(window,
                                           findContext,
                                           searchString,
                                           replaceString);
                    numReplacements += 1;
                    //Find_Replace(window,          /* editor */
                    //             findContext,     /* context */
                    //             searchString,    /* pattern */
                    //             replaceString);  /* replacement */
                }
            }
            var msg;
            if (numReplacements == 0) {
                msg = "'"+searchString+"' was not found in the specified range. No changes were made.";
            } else {
                msg = "Made "+numReplacements+" replacements in the specified range.";
            }
            window.StatusBar_AddMessage(msg, "find", 3000, true);
        }
    } catch (e) {
        vimlog.exception(e);
    }
}


VimController.special_command_mappings = {
    "edit"     : [ "cmd_revert" ] ,
    "help"     : [ "cmd_helpHelp " ],
    "next"     : [ "cmd_bufferNext" ] ,
    "previous" : [ "cmd_bufferPrevious" ] ,
    "quit"     : [ "cmd_bufferClose" ] ,
    "undo"     : [ "cmd_vim_undo" ] ,
    //"redo"     : [ "cmd_vim_redo" ] ,
    "write"    : [ "cmd_save" ] ,
    "wq"       : [ "cmd_save", "cmd_bufferClose" ] ,
    "xit"      : [ "cmd_save", "cmd_bufferClose" ] ,
    "exit"     : [ "cmd_save", "cmd_bufferClose" ]
};

// Regex for testing if this is a goto line command.
VimController.prototype.regexIsNumber = new RegExp('^\\d+$');

// Run command mode ":" string
VimController.prototype.findAndRunCommand = function (value)
{
    try {
        if (!value) {
            return;
        }
        // See if it's a number (goto line)
        if (this.regexIsNumber.test(value)) {
            var scimoz = gViewMgr.currentView.scintilla.scimoz;
            cmd_vim_gotoLine(scimoz, parseInt(value));
            return;
        } else if (value == '$') {
            // Not sure why, but had to use this function directly, maybe
            // something to with focus.
            //command_doCommand('cmd_documentEnd');
            var scimoz = gViewMgr.currentView.scintilla.scimoz;
            scimoz.documentEnd();
            return;
        }

        // See if it's !, which means it's a run command
        if (value[0] == '!') {
            value = value.substr(1);
            var userEnvSvc = Components.classes["@activestate.com/koUserEnviron;1"].
                  getService(Components.interfaces.koIUserEnviron);
            var env = userEnvSvc.GetEncodedEnvironment();
            Run_RunCommand(window,
                           value,
                           "%D",  // cwd (%D == directory path of file)
                           env,   // env
                           false, // insertOutput
                           false, // operateOnSelection
                           false, // doNotOpenOutputWindow
                           'command-output-window', // runIn
                           0,  // parseOutput
                           '', // parseRegex
                           0,  // showParsedOutputList
                           false); // saveInMRU
            return;
        }

        // See if it's a ranged command, begins with "x,y" or "%", where x and y
        // can be one of: number, "$" or "."
        // If it's not ranged, it means target the current line only.
        var scimoz = gViewMgr.currentView.scintilla.scimoz;
        // Initially, we only work on the current line
        var startLine = scimoz.lineFromPosition(scimoz.currentPos);
        var endLine = startLine;

        var rangeRegex = /^((\d+|\$|\.),(\d+|\$|\.)|%)/;
        var rangeRegexMatch = value.match(rangeRegex);
        if (rangeRegexMatch) {
            if (rangeRegexMatch[0] == "%") {
                // Everything
                startLine = 0;
                endLine   = scimoz.lineCount;
            } else {
                startLine = rangeRegexMatch[2];
                endLine   = rangeRegexMatch[3];
                // $ means end of file
                // . means current line
                if (startLine == '$') {
                    startLine = scimoz.lineCount;
                } else if (startLine == '.') {
                    startLine = scimoz.lineFromPosition(scimoz.currentPos);
                } else {
                    /* Should be a number then, convert to scintilla line */
                    startLine = startLine - 1;
                }
                if (endLine == '$') {
                    endLine = scimoz.lineCount;
                } else if (endLine == '.') {
                    endLine = scimoz.lineFromPosition(scimoz.currentPos);
                } else {
                    /* Should be a number then, convert to scintilla line */
                    endLine = endLine - 1;
                }
            }
            // We can now remove the range part from the command
            value = value.substr(rangeRegexMatch[0].length);
        }
        //dump("startLine: " + startLine + ", endLine: " + endLine + "\n");

        // See if it's a search and replace command
        var srMatch = true;
        var c; /* current char looking at */
        var buf = ''; /* stored buffer of read in chars */
        var fields = [];
        var state = 0;
        /* state == 0 ; then searching for 's' */
        /* state == 1 ; then getting searchString */
        /* state == 2 ; then getting replaceString */
        /* state == 3 ; then getting options */
        /* state == 4 ; then skipping escaped character */
        /* state > 4  ; then did not match */
        for (var i=0; srMatch && i < value.length; i++) {
            c = value[i];

            // Deal with slash characters
            if (c == '/' && state != 4) {
                state += 1;
                fields.push(buf);
                //dump("Field " + state + " is " + buf + "\n");
                buf = '';
                continue;
            }

            buf += c;
            if (c == '\\' && state != 4) {
                state = 4;
                continue;
            }

            switch (state) {
                case 0:
                    if (c != 's')
                        srMatch = false;
                    break;
                case 1:
                case 2:
                    break;
                case 3:
                    // Only accepts 'g', 'i', 'I' here. There are other options
                    // in vi though ('r', 'c', 'e', 'p', 'l', '#', 'n' ... ARGH!)
                    if (c != 'g' && c != 'i' && c != 'I')
                        srMatch = false;
                    break;
                case 4:
                    state = fields.length;
                    break;
                default:
                    srMatch = false;
                    break;
            }
        }
        if (srMatch) {
            if (buf.length > 0) {
                fields.push(buf);
            }
            if (fields.length >= 3 && fields.length <= 4) {
                //dump("Valid regex\n");
                //dump("fields (" + fields.length + "): " + fields + "\n");
                var currentPos = scimoz.currentPos;
                var options = '';
                if (fields.length == 4) {
                    options = fields[3];
                }
                scimoz.beginUndoAction();
                try {
                    this.performReplace(fields[1], fields[2], startLine, endLine, options);
                } catch (ex) {
                    vimlog.exception(e);
                }
                scimoz.endUndoAction();
                // Unselect the selection made
                scimoz.setSel(currentPos, currentPos);
                return;
            }
        }

        // No mapping, abbreviation or matched syntax existed!
        var commandName=value.match(/\w+/g);
        var forceCommand = false;
        if (commandName) {
            if (commandName.length < value.length) {
                if (value[commandName.length] == "!") {
                    forceCommand = true;
                }
            }
            //dump("Commandname: " + commandName + "\n");
            // Try exact match first
            var mappedCommands = VimController.special_command_mappings[value];
            if (!mappedCommands) {
                // Try an abbreviated command then
                for (var name in VimController.special_command_mappings) {
                    if ((name.length >= commandName.length) &&
                        (name.substr(0, commandName.length) == value)) {
                        mappedCommands = VimController.special_command_mappings[name];
                        break;
                    }
                }
            }
            if (mappedCommands) {
                // There is a mapping, perform the actions
                vimlog.debug("Performing command mode call for '" + mappedCommands + "'");
                var command;
                for (var i=0; i < mappedCommands.length; i++) {
                    command = mappedCommands[i];
                    vimlog.debug("Command mode called '" + command + "'");
                    command_doCommand(command);
                }
                return;
            }
        }

        // Parse the command, maybe there is a VI toolbox element to give it to
        var argsSplit = value.split(" ");
        var commandName = argsSplit.splice(0, 1);
        // Arguments - create them here
        // Lets go and check the toolbox for this name
        var viToolboxFolder = toolboxMgr.findPartByTypeAttributeValue('folder', 'name', 'Vi Commands');
        if (!!(viToolboxFolder)) {
            var part = viToolboxFolder.getChildByAttributeValue('name',
                                                                commandName,
                                                                true /* recurse */);
            if (part) {
                //dump("Value now: " + value + "\n");
                // Command details - create
                this._commandDetails.startLine = startLine;
                this._commandDetails.endLine = endLine;
                this._commandDetails.commandName = commandName;
                this._commandDetails.rawCommandString = value;
                this._commandDetails.setArguments(argsSplit.length, argsSplit);
                // Invoke the macro
                part_invokePart(part);
                // Command details - clear
                this._commandDetails.clear();
                return;
            }
        }

        dialog_alert("Unrecognized command: '" + commandName + "'\n\n" +
                     "You can implement additional commands by creating a toolbox folder\n" +
                     "named 'Vi Commands' and then creating a macro named '" + commandName + "'.\n" +
                     "This macro will then be executed in it's place.");
    } catch (e) {
        vimlog.exception(e);
    }
}




// Map vim commands to Komodo (scintilla) commands.
// These commands are a 1 to 1 mapping from vim to komodo (they don't do any
// thing special). We don't use a direct mapping from the key-bindings file as
// we want to be able to record all vim events. This mapping pass-through gives
// us the ability to run the commands as well as record what happened and thus
// enable special vim commands such as the repeat command.
VimController.command_mappings = {
// Movement actions
    "cmd_vim_left" :                [ VimController.SPECIAL_COMMAND,VimController.REPEATABLE_ACTION | VimController.MOVEMENT_ACTION | VimController.CHOOSE_CARET_X ],
    "cmd_vim_right" :               [ VimController.SPECIAL_COMMAND,VimController.REPEATABLE_ACTION | VimController.MOVEMENT_ACTION | VimController.CHOOSE_CARET_X ],
    "cmd_vim_linePrevious" :        [ "cmd_linePrevious",           VimController.REPEATABLE_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_lineNext" :            [ "cmd_lineNext",               VimController.REPEATABLE_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_linePreviousHome" :    [ "cmd_linePreviousHome",       VimController.REPEATABLE_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_lineNextHome" :        [ "cmd_lineNextHome",           VimController.REPEATABLE_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_home" :                [ "cmd_home",                   VimController.NO_REPEAT_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_homeAbsolute" :        [ "cmd_homeAbsolute",           VimController.NO_REPEAT_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_end"  :                [ "cmd_end",                    VimController.NO_REPEAT_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_pageDown" :            [ "cmd_pageDown",               VimController.REPEATABLE_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_pageUp" :              [ "cmd_pageUp",                 VimController.REPEATABLE_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_wordLeft" :            [ "cmd_wordLeft",               VimController.REPEATABLE_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_wordLeftEnd" :         [ "cmd_wordLeftEnd",            VimController.REPEATABLE_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_wordRight" :           [ "cmd_wordRight",              VimController.REPEATABLE_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_wordRightEnd" :        [ "cmd_wordRightEnd",           VimController.REPEATABLE_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_documentHome" :        [ "cmd_documentHome",           VimController.NO_REPEAT_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_documentEnd" :         [ "cmd_documentEnd",            VimController.NO_REPEAT_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_gotoLine" :            [ VimController.SPECIAL_COMMAND,VimController.NO_REPEAT_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_lineScrollUp" :        [ "cmd_lineScrollUp",           VimController.REPEATABLE_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_lineScrollDown" :      [ "cmd_lineScrollDown",         VimController.REPEATABLE_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_jumpToMatchingBrace" : [ "cmd_jumpToMatchingBrace",    VimController.NO_REPEAT_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_wordLeftPastPunctuation" : [ "cmd_wordLeftPastPunctuation", VimController.REPEATABLE_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_wordRightPastPunctuation" : [ "cmd_wordRightPastPunctuation", VimController.REPEATABLE_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_wordRightEndPastPunctuation" : [ "cmd_wordRightEndPastPunctuation", VimController.REPEATABLE_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_moveToScreenTop" :     [ "cmd_moveToScreenTop",        VimController.NO_REPEAT_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_moveToScreenCenter" :  [ "cmd_moveToScreenCenter",     VimController.NO_REPEAT_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_moveToScreenBottom" :  [ "cmd_moveToScreenBottom",     VimController.NO_REPEAT_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_moveSentenceBegin" :   [ "cmd_moveSentenceBegin",      VimController.REPEATABLE_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_moveSentenceEnd" :     [ "cmd_moveSentenceEnd",        VimController.REPEATABLE_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_moveParagraphBegin" :  [ "cmd_moveParagraphBegin",     VimController.REPEATABLE_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_moveParagraphEnd" :    [ "cmd_moveParagraphEnd",       VimController.REPEATABLE_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_moveFunctionPrevious" :[ "cmd_moveFunctionPrevious",   VimController.REPEATABLE_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_moveFunctionNext" :    [ "cmd_moveFunctionNext",       VimController.REPEATABLE_ACTION | VimController.MOVEMENT_ACTION ],
// Select actions
    "cmd_vim_selectCharNext" :      [ "cmd_selectCharNext",         VimController.REPEATABLE_ACTION ],
    "cmd_vim_selectCharPrevious" :  [ "cmd_selectCharPrevious",     VimController.REPEATABLE_ACTION ],
    "cmd_vim_selectWordLeft" :      [ "cmd_selectWordLeft",         VimController.REPEATABLE_ACTION ],
    "cmd_vim_selectWordRight" :     [ "cmd_selectWordRight",        VimController.REPEATABLE_ACTION ],
    "cmd_vim_selectHome" :          [ "cmd_selectHome",             VimController.NO_REPEAT_ACTION ],
    "cmd_vim_selectEnd" :           [ "cmd_selectEnd",              VimController.REPEATABLE_ACTION ],
    "cmd_vim_selectLineNext" :      [ "cmd_selectLineNext",         VimController.REPEATABLE_ACTION ],
    "cmd_vim_selectLinePrevious" :  [ "cmd_selectLinePrevious",     VimController.REPEATABLE_ACTION ],
    "cmd_vim_selectPageDown" :      [ "cmd_selectPageDown",         VimController.REPEATABLE_ACTION ],
    "cmd_vim_selectPageUp" :        [ "cmd_selectPageUp",           VimController.REPEATABLE_ACTION ],
    "cmd_vim_selectDocumentHome" :  [ "cmd_selectDocumentHome",     VimController.NO_REPEAT_ACTION ],
    "cmd_vim_selectDocumentEnd" :   [ "cmd_selectDocumentEnd",      VimController.NO_REPEAT_ACTION ],
    "cmd_vim_editSelectAll" :       [ "cmd_selectAll",              VimController.NO_REPEAT_ACTION ],
// Insert actions
    "cmd_vim_insert" :              [ VimController.SPECIAL_COMMAND,VimController.REPEATABLE_ACTION |
                                                                    VimController.ENTER_MODE_INSERT |
                                                                    VimController.MODIFY_ACTION |
                                                                    VimController.WORKS_IN_VISUAL_MODE |
                                                                    VimController.CANCELS_VISUAL_MODE ],
    "cmd_vim_insertHome" :          [ "cmd_home",                   VimController.REPEATABLE_ACTION |
                                                                    VimController.ENTER_MODE_INSERT |
                                                                    VimController.MODIFY_ACTION |
                                                                    VimController.WORKS_IN_VISUAL_MODE |
                                                                    VimController.CANCELS_VISUAL_MODE ],
    // XXX - Will not work on line 0, as we cannot do a linePrevious
    "cmd_vim_insert_newline_previous" : [ "cmd_newlinePrevious",    VimController.REPEATABLE_ACTION |
                                                                    VimController.ENTER_MODE_INSERT |
                                                                    VimController.MODIFY_ACTION ],
    "cmd_vim_insert_newline_next" : [ VimController.SPECIAL_COMMAND,VimController.REPEATABLE_ACTION |
                                                                    VimController.ENTER_MODE_INSERT |
                                                                    VimController.MODIFY_ACTION ],
    "cmd_vim_append" :              [ VimController.SPECIAL_COMMAND,VimController.REPEATABLE_ACTION |
                                                                    VimController.ENTER_MODE_INSERT |
                                                                    VimController.MODIFY_ACTION ],
    "cmd_vim_appendEnd" :           [ "cmd_end",                    VimController.REPEATABLE_ACTION |
                                                                    VimController.ENTER_MODE_INSERT |
                                                                    VimController.MODIFY_ACTION ],
    "cmd_vim_changeChar" :          [ VimController.SPECIAL_COMMAND,VimController.REPEATABLE_ACTION |
                                                                    VimController.ENTER_MODE_INSERT |
                                                                    VimController.MODIFY_ACTION |
                                                                    VimController.WORKS_IN_VISUAL_MODE |
                                                                    VimController.CANCELS_VISUAL_MODE ],
    "cmd_vim_changeWord" :          [ VimController.SPECIAL_COMMAND,VimController.REPEATABLE_ACTION |
                                                                    VimController.ENTER_MODE_INSERT |
                                                                    VimController.MODIFY_ACTION ],
    "cmd_vim_changeLine" :          [ VimController.SPECIAL_COMMAND,VimController.REPEATABLE_ACTION |
                                                                    VimController.ENTER_MODE_INSERT |
                                                                    VimController.MODIFY_ACTION |
                                                                    VimController.SPECIAL_REPEAT_HANDLING ],
    "cmd_vim_changeLineEnd" :       [ VimController.SPECIAL_COMMAND,VimController.NO_REPEAT_ACTION |
                                                                    VimController.ENTER_MODE_INSERT |
                                                                    VimController.MODIFY_ACTION |
                                                                    VimController.WORKS_IN_VISUAL_MODE |
                                                                    VimController.CANCELS_VISUAL_MODE ],
// Cut/copy/delete actions
    "cmd_vim_yankLine" :            [ VimController.SPECIAL_COMMAND,VimController.REPEATABLE_ACTION |
                                                                    VimController.COPY_LINES |
                                                                    VimController.SPECIAL_REPEAT_HANDLING ],
    // Paste actions are special because of the way vim uses yanked lines
    "cmd_vim_paste" :               [ VimController.SPECIAL_COMMAND,VimController.REPEATABLE_ACTION |
                                                                    VimController.MODIFY_ACTION |
                                                                    VimController.WORKS_IN_VISUAL_MODE |
                                                                    VimController.CANCELS_VISUAL_MODE ],
    "cmd_vim_pasteAfter" :          [ VimController.SPECIAL_COMMAND,VimController.REPEATABLE_ACTION |
                                                                    VimController.MODIFY_ACTION |
                                                                    VimController.WORKS_IN_VISUAL_MODE |
                                                                    VimController.CANCELS_VISUAL_MODE ],
    "cmd_vim_lineCut" :             [ VimController.SPECIAL_COMMAND,VimController.REPEATABLE_ACTION |
                                                                    VimController.MODIFY_ACTION |
                                                                    VimController.COPY_LINES |
                                                                    VimController.WORKS_IN_VISUAL_MODE |
                                                                    VimController.CANCELS_VISUAL_MODE ],
    // Same "cmd_vim_deleteLineToEnd" :     [ "cmd_deleteLineToEnd",                   VimController.REPEATABLE_ACTION ],
    "cmd_vim_lineCutEnd" :          [ VimController.SPECIAL_COMMAND,VimController.REPEATABLE_ACTION |
                                                                    VimController.MODIFY_ACTION |
                                                                    VimController.COPY_NORMAL |
                                                                    VimController.WORKS_IN_VISUAL_MODE |
                                                                    VimController.CANCELS_VISUAL_MODE ],
    "cmd_vim_cutChar" :             [ VimController.SPECIAL_COMMAND,VimController.REPEATABLE_ACTION |
                                                                    VimController.MODIFY_ACTION |
                                                                    VimController.COPY_NORMAL |
                                                                    VimController.WORKS_IN_VISUAL_MODE |
                                                                    VimController.CANCELS_VISUAL_MODE ],
    "cmd_vim_cutCharLeft" :         [ VimController.SPECIAL_COMMAND,VimController.REPEATABLE_ACTION ],
    "cmd_vim_yankOperation": [ VimController.SPECIAL_COMMAND,       VimController.NO_REPEAT_ACTION |
                                                                    VimController.COPY_NORMAL |
                                                                    VimController.WORKS_IN_VISUAL_MODE |
                                                                    VimController.CANCELS_VISUAL_MODE ],
    "cmd_vim_deleteOperation": [ VimController.SPECIAL_COMMAND,     VimController.NO_REPEAT_ACTION |
                                                                    VimController.COPY_NORMAL |
                                                                    VimController.WORKS_IN_VISUAL_MODE |
                                                                    VimController.CANCELS_VISUAL_MODE ],
    "cmd_vim_changeOperation": [ VimController.SPECIAL_COMMAND,     VimController.NO_REPEAT_ACTION |
                                                                    VimController.COPY_NORMAL |
                                                                    VimController.WORKS_IN_VISUAL_MODE |
                                                                    VimController.CANCELS_VISUAL_MODE ],
// Search actions
    "cmd_startIncrementalSearch" :  [ "cmd_startIncrementalSearch", VimController.NO_REPEAT_ACTION ],
    "cmd_vim_findNext" :            [ VimController.SPECIAL_COMMAND,VimController.NO_REPEAT_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_findPrevious" :        [ VimController.SPECIAL_COMMAND,VimController.NO_REPEAT_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_findNextResult" :      [ "cmd_findNextResult",         VimController.NO_REPEAT_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_findCharInLine" :      [ VimController.SPECIAL_COMMAND,VimController.NO_REPEAT_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_findPreviousCharInLine":[VimController.SPECIAL_COMMAND,VimController.NO_REPEAT_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_findWordUnderCursor" : [ VimController.SPECIAL_COMMAND,VimController.NO_REPEAT_ACTION | VimController.MOVEMENT_ACTION ],
    "cmd_vim_findWordUnderCursorBack" : [ VimController.SPECIAL_COMMAND,VimController.NO_REPEAT_ACTION | VimController.MOVEMENT_ACTION ],
// Visual mode commands
    "cmd_vim_toggleVisualMode" :    [ VimController.SPECIAL_COMMAND,VimController.NO_REPEAT_ACTION | VimController.WORKS_IN_VISUAL_MODE ],
    "cmd_vim_toggleVisualLineMode" :[ VimController.SPECIAL_COMMAND,VimController.NO_REPEAT_ACTION | VimController.WORKS_IN_VISUAL_MODE ],
    "cmd_vim_toggleVisualBlockMode" : [ VimController.SPECIAL_COMMAND,VimController.NO_REPEAT_ACTION | VimController.WORKS_IN_VISUAL_MODE ],
// Special actions
    "cmd_vim_replaceChar" :         [ VimController.SPECIAL_COMMAND,VimController.REPEATABLE_ACTION | VimController.MODIFY_ACTION |
                                                                    VimController.WORKS_IN_VISUAL_MODE |
                                                                    VimController.CANCELS_VISUAL_MODE ],
    "cmd_vim_overtype" :            [ VimController.SPECIAL_COMMAND,VimController.NO_REPEAT_ACTION ],
    "cmd_vim_swapCase" :            [ "cmd_swapCase",               VimController.REPEATABLE_ACTION | VimController.MODIFY_ACTION |
                                                                    VimController.WORKS_IN_VISUAL_MODE |
                                                                    VimController.CANCELS_VISUAL_MODE ],
    "cmd_vim_join" :                [ "cmd_join",                   VimController.REPEATABLE_ACTION | VimController.MODIFY_ACTION |
                                                                    VimController.WORKS_IN_VISUAL_MODE |
                                                                    VimController.CANCELS_VISUAL_MODE ],
    "cmd_vim_undo" :                [ "cmd_undo",                   VimController.NO_REPEAT_ACTION ],
    "cmd_vim_redo" :                [ "cmd_redo",                   VimController.NO_REPEAT_ACTION ],
    "cmd_vim_cancel" :              [ VimController.SPECIAL_COMMAND,VimController.NO_REPEAT_ACTION |
                                                                    VimController.WORKS_IN_VISUAL_MODE |
                                                                    VimController.CANCELS_VISUAL_MODE ],
    "cmd_vim_repeatLastCommand" :   [ VimController.SPECIAL_COMMAND,VimController.REPEATABLE_ACTION ],
    "cmd_vim_enterCommandMode" :    [ VimController.SPECIAL_COMMAND,VimController.NO_REPEAT_ACTION |
                                                                    VimController.WORKS_IN_VISUAL_MODE ],
    "cmd_vim_dedent" :              [ VimController.SPECIAL_COMMAND,VimController.REPEATABLE_ACTION | VimController.MODIFY_ACTION |
                                                                    VimController.WORKS_IN_VISUAL_MODE |
                                                                    VimController.CANCELS_VISUAL_MODE ],
    "cmd_vim_indent" :              [ VimController.SPECIAL_COMMAND,VimController.REPEATABLE_ACTION | VimController.MODIFY_ACTION |
                                                                    VimController.WORKS_IN_VISUAL_MODE |
                                                                    VimController.CANCELS_VISUAL_MODE ],
    "cmd_vim_enterSearchForward" :  [ VimController.SPECIAL_COMMAND,VimController.NO_REPEAT_ACTION |
                                                                    VimController.WORKS_IN_VISUAL_MODE ],
    "cmd_vim_enterSearchBackward" : [ VimController.SPECIAL_COMMAND,VimController.NO_REPEAT_ACTION |
                                                                    VimController.WORKS_IN_VISUAL_MODE ],
    "" :                            [ "",                           VimController.NO_REPEAT_ACTION ]
};

// Special operation commands, used in conjunction with a movement command, or
// if repeated twice (like "d d") then performs the associated action below.
VimController.operation_commands = {
    "cmd_vim_yankOperation": [ "cmd_vim_yankLine", VimController.OPERATION_YANK ],
    "cmd_vim_deleteOperation": [ "cmd_vim_lineCut", VimController.OPERATION_DELETE ],
    "cmd_vim_changeOperation": [ "cmd_vim_changeLine", VimController.OPERATION_CHANGE ]
}

// Visual mode commands that are different to normal mode commands.
VimController.visualmode_command_mappings = {
    "cmd_vim_cutChar":          "cmd_vim_deleteOperation",
    "cmd_vim_changeChar":       "cmd_vim_insert",
    // Left and right are handled locally as they act differently
    //"cmd_vim_left":             "cmd_vim_selectCharPrevious",
    //"cmd_vim_right":            "cmd_vim_selectCharNext",
    "cmd_vim_wordLeft" :        "cmd_vim_selectWordLeft",
    "cmd_vim_wordRight" :       "cmd_vim_selectWordRight",
    "cmd_vim_home" :            "cmd_vim_selectHome",
    // XXX - Implement 'cmd_vim_selectHomeAbsolute'
    //"cmd_vim_homeAbsolute" :    "cmd_vim_selectHomeAbsolute",
    "cmd_vim_end" :             "cmd_vim_selectEnd",
    "cmd_vim_lineNext" :        "cmd_vim_selectLineNext",
    "cmd_vim_linePrevious" :    "cmd_vim_selectLinePrevious",
    "cmd_vim_pageDown" :        "cmd_vim_selectPageDown",
    "cmd_vim_pageUp" :          "cmd_vim_selectPageUp",
    //"cmd_vim_documentHome" :    "cmd_vim_selectDocumentHome",
    "cmd_vim_documentEnd" :     "cmd_vim_selectDocumentEnd",
    "": ""
}

// Execute a vim command.
function vim_doCommand(command)
{
    vimlog.debug("vim_doCommand(" + command + ")");
    var rc = false;
    var original_command = command;
    try {
        var numRepeats = 1;

        // Must have a 1-to-1 mapping in the dict to be a komodo vi command.
        var mappedCommand = VimController.command_mappings[command];
        if (!(!!mappedCommand)) {
            // No mapping existed
            vimlog.error("vim_doCommand:: Unknown vim command: '" + command + "'");
            gVimController.resetCommandSettings();
            return false;
        }

        var movementAction = mappedCommand[1] & VimController.MOVEMENT_ACTION;
        var scimoz = gViewMgr.currentView.scintilla.scimoz;
        // Record current position and selection (anchor is used for selection)
        var orig_currentPos = scimoz.currentPos;
        var orig_anchor = scimoz.anchor;
        gVimController._currentPos = orig_currentPos
        gVimController._anchor = orig_anchor;

        // Visual mode checks
        if (gVimController.mode == VimController.MODE_VISUAL) {
            // If in visual mode, the only things we want to see are movements
            // or special visual mode commands
            if (!movementAction &&
                !(mappedCommand[1] & VimController.WORKS_IN_VISUAL_MODE)) {
                vimlog.debug("vim_doCommand:: visual mode:: command not recogized: " + command);
                window.StatusBar_AddMessage("Vi: Visual Mode: not a recognized visual mode command",
                                            "vimode", 3000, false);
                gVimController.resetCommandSettings();
                // Reset the repeat count back to 0 now
                gVimController.repeatCount = 0;
                return false;
            }

            // See if we need to change how the command works
            if (command in VimController.visualmode_command_mappings) {
                command = VimController.visualmode_command_mappings[command];
                vimlog.debug("vim_doCommand:: visual mode:: command mapped to: " + command);
                mappedCommand = VimController.command_mappings[command];
                movementAction = mappedCommand[1] & VimController.MOVEMENT_ACTION;
            }
        }

        // See if this is a operation that requires a movement command
        //   "d G", "c w", etc...
        var opMovementAction = false;
        var original_operationFlags = gVimController.operationFlags;
        vimlog.debug("vim_doCommand:: operationFlags: "+original_operationFlags);
        if (gVimController.operationFlags) {
            // Check if the same command is repeated, then it's a special case
            if ((command in VimController.operation_commands) &&
                (gVimController.operationFlags == VimController.operation_commands[command][1])) {
                command = VimController.operation_commands[command][0];
                vimlog.debug("vim_doCommand:: operation command repeated, changing to: " + command);
                mappedCommand = VimController.command_mappings[command];
            } else {
                // If it's not a movement command or the same command again
                if (!movementAction) {
                    vimlog.debug("vim_doCommand:: operation not in conjuction with a movement");
                    window.StatusBar_AddMessage("Vi: "+VimController.OPERATION_NAMES[gVimController.operationFlags]+
                                                " operation must occur in conjunction with a movement command",
                                                "vimode", 3000, false);
                    gVimController.resetCommandSettings();
                    return false;
                }
                vimlog.debug("vim_doCommand:: is a opMovementAction");
                opMovementAction = true;
                var currentPos = scimoz.currentPos;
                if ((command == 'cmd_vim_wordRight') &&
                    (gVimController.operationFlags == VimController.OPERATION_CHANGE)) {
                    // Special case command, does not eat whitespace like cmd_wordRight
                    vimlog.debug("Special casing cmd_vim_wordRight for change command, mapping to cmd_vim_changeWord instead");
                    command = 'cmd_vim_changeWord';
                    mappedCommand = VimController.command_mappings[command];
                }
            }
        }

        // There is a mapping, if repeatable, record this action
        // Only repeatable actions will be able to loop n times
        if (mappedCommand[1] & VimController.REPEATABLE_ACTION) {
            gVimController._lastOperationFlags = gVimController.operationFlags;
            if (gVimController.repeatCount > 1) {
                numRepeats = gVimController.repeatCount;
                vimlog.debug("vim_doCommand:: Repeat command: " + numRepeats + " times");
            }
            // Reset the repeat count back to 0 now
            gVimController.repeatCount = 0;
        }

        // For copying, record what type of copy action it is
        if (mappedCommand[1] & VimController.COPY_ACTION) {
            gVimController._copyMode = mappedCommand[1] & VimController.COPY_ACTION;
            // reset the internal copy buffer
            gVimController._internalBuffer = "";
        }

        var commandName = mappedCommand[0];   // The command name
        if (commandName != VimController.SPECIAL_COMMAND) {
            // Set the command to the mapped command
            command = commandName;
        // } else {
            // Use it's own special handler in this file cmd_vim_...
        }

        // Record modifying action if makes modifications
        var modifyAction = (mappedCommand[1] & VimController.MODIFY_ACTION ||
                            gVimController.operationFlags & VimController.OPERATION_DELETE);
        if (modifyAction) {
            gVimController._lastModifyCommand = command;
            gVimController._lastRepeatCount = numRepeats;
            gVimController._lastOperationFlags = gVimController.operationFlags;
            gVimController._lastInsertedKeycodes = [];
        }

        // XXX - Leaving limitation of not knowing undo/redo history repitition count
        //if (gVimController._commandHistory.length > gVimController._commandHistoryPosition) {
        //    gVimController._commandHistory = gVimController._commandHistory.slice(0, gVimController._commandHistory.length);
        //}
        //gVimController._commandHistoryPosition = gVimController._commandHistory.length;
        //gVimController._commandHistory.push(mappedCommand);
        //gVimController._commandHistoryRepeats.push(numRepeats);

        // For modify actions, Wrap the whole vi command in an undo action
        if (modifyAction) {
            //dump("Vi beginUndoAction\n");
            scimoz.beginUndoAction();
        }
        try {
            if (commandName == VimController.SPECIAL_COMMAND) {
                var recordingMacro = false;
                // Handle macro recording: We turn off macro recording whilst
                // we perform the commands, as some may make calls to the
                // do_command() function, which will also record macros and we
                // don't want these recorded twice.
                if (typeof(gMacroMgr) != 'undefined' && gMacroMgr.mode == 'recording') {
                    recordingMacro = true;
                    for (var i = 0; i < numRepeats; i++) {
                        gMacroMgr.appendCommand(command);
                    }
                    gMacroMgr.mode = 'stopped';
                }
                // Local to this javascript file, call the function directly
                if (mappedCommand[1] & VimController.SPECIAL_REPEAT_HANDLING) {
                    // Uses repeat count specially
                    vimlog.debug("vim_doCommand:: Special repeat handling used");
                    eval(command + "(scimoz, " + numRepeats + ")");
                } else {
                    for (var i = 0; i < numRepeats; i++) {
                        eval(command + "(scimoz)");
                    }
                }
                if (recordingMacro) {
                    gMacroMgr.mode = 'recording';
                }
            } else {
                // Let the controllers find the necessary command
                for (var i = 0; i < numRepeats; i++) {
                    command_doCommand(command);
                }
            }

            if (opMovementAction) {
                var movePos = scimoz.currentPos;
                // Ensure start offset for copyInternal is smaller than the
                // end offset.
                vimlog.debug("vim_doCommand:: opFlags:"+gVimController.operationFlags+
                          ", copying buffer between: "+currentPos+
                          ", movePos:"+movePos);
                if (movePos < currentPos) {
                    gVimController.copyInternal(scimoz, movePos, currentPos,
                                                modifyAction);
                } else {
                    gVimController.copyInternal(scimoz, currentPos, movePos,
                                                modifyAction);
                }
                // If we are just yanking, the cursor does not move.
                if (gVimController.operationFlags == VimController.OPERATION_YANK) {
                    vimlog.debug("vim_doCommand:: OPERATION_YANK:: moving cursor back to starting position");
                    scimoz.currentPos = currentPos;
                    scimoz.anchor = currentPos;
                }
            }
        } catch (ex) {
            vimlog.error("vim_doCommand:: Command '"+original_command+"' failed:");
            vimlog.exception(ex);
        }
        if (modifyAction) {
            //dump("Vi endUndoAction\n");
            scimoz.endUndoAction();
        }

        if (command in VimController.operation_commands) {
            // Check if it's in visual mode or we have selected text
            var selectionStart = scimoz.anchor;
            var selectionEnd = scimoz.currentPos;
            if (gVimController.mode == VimController.MODE_VISUAL) {
                if (gVimController._visualMode == VimController.VISUAL_LINE) {
                    // Visual line mode uses different settings from scimoz API
                    if (selectionStart < selectionEnd) {
                        selectionStart = scimoz.getLineSelStartPosition(scimoz.lineFromPosition(selectionStart));
                        selectionEnd = scimoz.getLineSelEndPosition(scimoz.lineFromPosition(selectionEnd));
                    } else {
                        selectionEnd = scimoz.getLineSelStartPosition(scimoz.lineFromPosition(selectionEnd));
                        selectionStart = scimoz.getLineSelEndPosition(scimoz.lineFromPosition(selectionStart));
                    }
                    // Set copyMode to line type, so pastes will work properly
                    gVimController._copyMode = VimController.COPY_LINES;
                } else if (selectionStart == selectionEnd) {
                    // Visual mode, but no selection, operate on char at currentPos
                    selectionStart = scimoz.positionAfter(selectionEnd);
                }
            }
            if (selectionStart != selectionEnd) {
                vimlog.debug("vim_doCommand:: opFlags:"+gVimController.operationFlags+
                          ", copying buffer between: "+selectionStart+
                          " and "+selectionEnd);
                if (selectionStart < selectionEnd) {
                    gVimController.copyInternal(scimoz, selectionStart, selectionEnd,
                                                gVimController.operationFlags & VimController.OPERATION_DELETE);
                    // Move to the start of the selection after the copy/delete
                    gVimController._currentPos = selectionStart;
                } else {
                    gVimController.copyInternal(scimoz, selectionEnd, selectionStart,
                                                gVimController.operationFlags & VimController.OPERATION_DELETE);
                }
                gVimController.resetCommandSettings();
            }
        } else {
            // reset the operationFlags back to normal.
            gVimController.resetCommandSettings();
        }

        // If we are still in visual mode, update the selection or leave
        // visual mode for special cased commands.
        if (gVimController.mode == VimController.MODE_VISUAL) {
            if (mappedCommand[1] & VimController.CANCELS_VISUAL_MODE) {
                gVimController.mode = VimController.MODE_NORMAL;
            }
        }

        // Enter insert mode if the command specified it or if it's a
        // change text operation (like "c w")
        if ((mappedCommand[1] & VimController.ENTER_MODE_INSERT) ||
            (opMovementAction && (gVimController._lastOperationFlags == VimController.OPERATION_CHANGE))) {
            gVimController.mode = VimController.MODE_INSERT;
            // Start recording what gets entered
            gVimController._lastInsertedKeycodes = [];
        //} else {
        //    // Go back to normal mode
        //    gVimController.mode = VimController.MODE_NORMAL;
        }

        var new_currentPos = gVimController._currentPos;
        if (new_currentPos == orig_currentPos) {
            new_currentPos = null;  /* selection handler will work it out */
        }
        gVimController.updateCursorAndSelection(scimoz,
                                                orig_currentPos,
                                                new_currentPos,
                                                orig_anchor,
                                                mappedCommand[1] & VimController.CHOOSE_CARET_X);

        rc = true;
    } catch (e) {
        vimlog.exception(e);
        vimlog.error("vim_doCommand:: An error occurred executing the command '" + command + "'");
        throw Components.results.NS_ERROR_FAILURE;
    }
    return rc;
}


function cmd_vim_repeatLastCommand(scimoz) {
    try {
        if (gVimController._lastModifyCommand) {
            vimlog.debug("cmd_vim_repeatLastCommand: _lastModifyCommand: " + gVimController._lastModifyCommand);
            vimlog.debug("cmd_vim_repeatLastCommand: repeatCount: " + gVimController.repeatCount);
            vimlog.debug("cmd_vim_repeatLastCommand: _lastRepeatCount: " + gVimController._lastRepeatCount);
            vimlog.debug("cmd_vim_repeatLastCommand: _lastOperationFlags: " + gVimController._lastOperationFlags);
            vimlog.debug("cmd_vim_repeatLastCommand: _lastInsertedKeycodes: " + gVimController._lastInsertedKeycodes.join(""));

            // Work out how many times to perform the command
            var numRepeats = 1;
            if (gVimController.repeatCount > 1) {
                numRepeats = gVimController.repeatCount;
            }

            // The idea here is convert the keys to one block of text and to
            // insert that text into scintilla directly.
            var textLines = [];
            var keys = gVimController._lastInsertedKeycodes;
            if (keys) {
                var textPos = 0;
                var textArray;
                var text;
                var i;
                for (i=0; i < keys.length; i++) {
                    if (keys[i] == "\n") {
                        textArray = keys.slice(textPos, i);
                        text = textArray.join("");
                        textLines.push(text);
                        textPos = i + 1;
                    }
                }
                textArray = keys.slice(textPos, i);
                text = textArray.join("");
                textLines.push(text);
            }

            // Make a copy of these, as they will get modified as we run
            // commands, then restore them once were done. Bug 58627
            var lastRepeatCount = Math.max(1, gVimController._lastRepeatCount);
            var lastModifyCommand = gVimController._lastModifyCommand;
            var lastOperationFlags = gVimController._lastOperationFlags;

            // Now we perform the repeat action
            //dump("Vi beginUndoAction\n");
            scimoz.beginUndoAction();
            try {
                var i, j;
                for (i=0; i < numRepeats; i++) {
                    for (j=0; j < lastRepeatCount; j++) {
                        command_doCommand(lastModifyCommand);
                    }
                    // Re-add the text if there is some
                    for (j=0; j < textLines.length; j++) {
                        scimoz.addText(stringutils_sysUtils.byteLength(textLines[j]), textLines[j]);
                        if (j < (textLines.length - 1)) {
                            // Don't do this for the last block of text
                            command_doCommand('cmd_newline');
                        }
                    }
                }
            } catch (ex) {
            }
            //dump("Vi endUndoAction\n");
            scimoz.endUndoAction();

            // Reset the state back to what it was like before the command
            gVimController.mode = VimController.MODE_NORMAL;
            gVimController._lastInsertedKeycodes = keys;
            gVimController._lastRepeatCount = lastRepeatCount;
            gVimController._lastModifyCommand = lastModifyCommand;
            gVimController._lastOperationFlags = lastOperationFlags;
        }
    } catch (e) {
        vimlog.exception(e);
    }
}

// Movement commands
function cmd_vim_left(scimoz) {
    // Cannot move past the beginning of the current line
    var lineNo = scimoz.lineFromPosition(gVimController._currentPos);
    var lineStartPos = scimoz.positionFromLine(lineNo);
    if (gVimController._currentPos > lineStartPos) {
        gVimController._currentPos = scimoz.positionBefore(gVimController._currentPos);
    }
}

function cmd_vim_right(scimoz) {
    // Cannot move past the end of the current line
    var lineNo = scimoz.lineFromPosition(gVimController._currentPos);
    var lineEndPos = scimoz.getLineEndPosition(lineNo);
    if (gVimController._currentPos < lineEndPos) {
        gVimController._currentPos = scimoz.positionAfter(gVimController._currentPos);
    }
}

function cmd_vim_gotoLine(scimoz, lineNumber) {
    try {
        if (typeof(lineNumber) != 'undefined') {
            scimoz.gotoLine(lineNumber - 1);  // scimoz handles out of bounds
        } else if (gVimController.repeatCount > 0) {
            scimoz.gotoLine(gVimController.repeatCount - 1);  // scimoz handles out of bounds
            gVimController.repeatCount = 0;
        } else {
            command_doCommand('cmd_documentEnd');
        }
    } catch (e) {
        vimlog.exception(e);
    }
}

function cmd_vim_insert(scimoz) {
    scimoz.replaceSel("");
}

function cmd_vim_append(scimoz) {
    try {
        // We don't want to go past the end of the line
        var currentPos = scimoz.currentPos;
        var lineNo = scimoz.lineFromPosition(currentPos);
        var lineEndPos = scimoz.getLineEndPosition(lineNo)
        if (currentPos < lineEndPos) {
            command_doCommand('cmd_right');
        }
    } catch (e) {
        vimlog.exception(e);
    }
}

function cmd_vim_changeChar(scimoz) {
    try {
        cmd_vim_cutChar(scimoz);
    } catch (e) {
        vimlog.exception(e);
    }
}

// This is a little bit special, as we don't want to eat the whitespace after
// the word, which is what cmd_wordRight does.
function cmd_vim_changeWord(scimoz) {
    try {
        var currentPos = scimoz.currentPos;
        var end = scimoz.wordEndPosition(currentPos, false);
        gVimController.copyInternal(scimoz, currentPos, end, true);
    } catch (e) {
        vimlog.exception(e);
    }
}

function cmd_vim_changeLine(scimoz, repeatCount) {
    try {
        if (!repeatCount) {
            repeatCount = 1;
        }
        // We don't actually delete all of the last line, just till EOL
        repeatCount -= 1;
        var lineNo = scimoz.lineFromPosition(gVimController._currentPos);
        var start = scimoz.positionFromLine(lineNo);
        var endLine = Math.min(lineNo+repeatCount, scimoz.lineCount);
        var end = scimoz.getLineEndPosition(endLine);
        gVimController.copyInternal(scimoz, start, end, true);
    } catch (e) {
        vimlog.exception(e);
    }
}

function cmd_vim_yankLine(scimoz, repeatCount) {
    try {
        if (!repeatCount) {
            repeatCount = 1;
        }
        var currentPos = scimoz.currentPos;
        var lineNo = scimoz.lineFromPosition(currentPos);
        var start = scimoz.positionFromLine(lineNo);
        var endLine = Math.min(lineNo+repeatCount, scimoz.lineCount);
        var end = scimoz.positionFromLine(endLine);
        gVimController.copyInternal(scimoz, start, end);
    } catch (e) {
        vimlog.exception(e);
    }
}

function cmd_vim_changeLineEnd(scimoz) {
    cmd_vim_lineCutEnd(scimoz);
}

function cmd_vim_cutChar(scimoz) {
    try {
        var currentPos = scimoz.currentPos;
        var lineNo = scimoz.lineFromPosition(currentPos);
        var lineEndPos = scimoz.getLineEndPosition(lineNo);
        if (currentPos < lineEndPos) {
            // delete character to the right of the cursor
            var end = scimoz.positionAfter(currentPos);
            gVimController.copyInternal(scimoz, currentPos, end, true);
        }
    } catch (e) {
        vimlog.exception(e);
    }
}

function cmd_vim_cutCharLeft(scimoz) {
    try {
        var currentPos = scimoz.currentPos;
        var lineNo = scimoz.lineFromPosition(currentPos);
        var lineStartPos = scimoz.positionFromLine(lineNo);
        if (currentPos > lineStartPos) {
            // delete character to the right of the cursor
            var start = scimoz.positionBefore(currentPos);
            gVimController.copyInternal(scimoz, start, currentPos, true);
        }
    } catch (e) {
        vimlog.exception(e);
    }
}

function cmd_vim_deleteToDocumentEnd(scimoz) {
    try {
        gVimController.copyInternal(scimoz, scimoz.currentPos, scimoz.length,
                                    true);
    } catch (e) {
        vimlog.exception(e);
    }
}

/**
 * Start yank operation, ie: "y w", "y e", "y G", etc...
 * Have to wait for till after the movement command to see how much to yank.
 */
function cmd_vim_yankOperation(scimoz) {
    vimlog.debug("cmd_vim_yankOperation:: Setting to yank operation");
    gVimController.operationFlags = VimController.OPERATION_YANK;
}

/**
 * Start delete operation, ie: "d w", "d e", "d G", etc...
 * Have to wait for till after the movement command to see how much to delete.
 */
function cmd_vim_deleteOperation(scimoz) {
    if (gVimController.operationFlags == VimController.OPERATION_DELETE) {
        // "d d" means to delete the current line
        cmd_vim_lineCut(scimoz);
    } else {
        gVimController.operationFlags = VimController.OPERATION_DELETE;
    }
}

/**
 * Start change operation, ie: "c w", "c e", "c G", etc...
 * Have to wait for till after the movement command to see how much to change.
 */
function cmd_vim_changeOperation(scimoz) {
    gVimController.operationFlags = VimController.OPERATION_CHANGE;
}

function cmd_vim_insert_newline_next(scimoz) {
    command_doCommand('cmd_end');
    command_doCommand('cmd_newline');
}

function cmd_vim_lineCut(scimoz) {
    try {
        var currentPos = scimoz.currentPos;
        var lineNo = scimoz.lineFromPosition(currentPos);
        if (lineNo < scimoz.lineCount) {
            var start = scimoz.positionFromLine(lineNo);
            var end = scimoz.positionFromLine(lineNo+1);
            gVimController.copyInternal(scimoz, start, end, true);
        }
    } catch (e) {
        vimlog.exception(e);
    }
}

function cmd_vim_lineCutEnd(scimoz) {
    try {
        var currentPos = scimoz.currentPos;
        var lineNo = scimoz.lineFromPosition(currentPos);
        var end = scimoz.getLineEndPosition(lineNo);
        gVimController.copyInternal(scimoz, currentPos, end, true);
    } catch (e) {
        vimlog.exception(e);
    }
}

function _cmd_vim_pasteLines_handler(scimoz, pasteAfter) {
    try {
        var currentPos = scimoz.currentPos;
        // Go to column 0 on current line
        if (pasteAfter) {
            var lineNo = scimoz.lineFromPosition(currentPos);
            if (lineNo == scimoz.lineCount) {
                // Case where already on the last line (lineDown no help)
                scimoz.newLine();
            } else {
                scimoz.lineDown();
            }
        }
        scimoz.home();
        var start = scimoz.currentPos;
        var buf = gVimController._internalBuffer;
        // Any selection is removed as part of the paste
        scimoz.replaceSel("");
        // Must use a byte length here, see bug 63498
        scimoz.addText(stringutils_sysUtils.byteLength(buf), buf);
        scimoz.anchor = scimoz.currentPos;
        scimoz.currentPos = start;
        scimoz.vCHomeWrap();
    } catch (e) {
        vimlog.exception(e);
    }
}

function _cmd_vim_paste(scimoz) {
    try {
        var currentPos = scimoz.currentPos;
        var start = scimoz.currentPos;
        var buf = gVimController._internalBuffer;
        // Any selection is removed as part of the paste
        scimoz.replaceSel("");
        scimoz.addText(stringutils_sysUtils.byteLength(buf), buf);
    } catch (e) {
        vimlog.exception(e);
    }
}

function cmd_vim_paste(scimoz) {
    if (gVimController._copyMode == VimController.COPY_LINES) {
        // Special case for yanked lines
        _cmd_vim_pasteLines_handler(scimoz, false);
    } else {
        _cmd_vim_paste(scimoz);
    }
}

function cmd_vim_pasteAfter(scimoz) {
    if (gVimController._copyMode == VimController.COPY_LINES) {
        // Special case for yanked lines
        _cmd_vim_pasteLines_handler(scimoz, true);
    } else {
        vim_doCommand('cmd_vim_right');
        _cmd_vim_paste(scimoz);
    }
}

function cmd_vim_cancel(scimoz) {
    if (gVimController.mode == VimController.MODE_COMMAND) {
        gVimController.inputBufferFinish();
    } else if (gVimController.mode == VimController.MODE_INSERT) {
        // Normal vi will move the cursor one to the left here
        // XXX TODO - We may have a repeat count, perform multiples if needed.
        var lineNo = scimoz.lineFromPosition(gVimController._currentPos);
        var lineStartPos = scimoz.positionFromLine(lineNo);
        if (gVimController._currentPos > lineStartPos) {
            //command_doCommand('cmd_left');
            gVimController._currentPos = scimoz.positionBefore(gVimController._currentPos);
        }
    }
    gVimController.mode = VimController.MODE_NORMAL;
    gVimController.repeatCount = 0;
    gVimController.operationFlags = VimController.OPERATION_NONE;
}

function cmd_vim_findNext(scimoz) {
    gVimController.performSearch(scimoz, null);
}

function cmd_vim_findPrevious(scimoz) {
    gVimController.performSearch(scimoz, null, true /* reverseDirection */);
}

function cmd_vim_findCharInLine(scimoz) {
    // Next character typed, we'll search for it
    gVimController.mode = VimController.MODE_FIND_CHAR;
    gVimController._findCharDirection = VimController.SEARCH_FORWARD;
}

function cmd_vim_findPreviousCharInLine(scimoz) {
    // Next character typed, we'll search for it backwards
    gVimController.mode = VimController.MODE_FIND_CHAR;
    gVimController._findCharDirection = VimController.SEARCH_BACKWARD;
}

function _cmd_vim_findWordUnderCursor_wrapper(scimoz, searchDirection) {
    try {
        var currentPos = scimoz.currentPos;
        var lineNo = scimoz.lineFromPosition(currentPos);
        var lineStartPos = scimoz.positionFromLine(lineNo);
        var lineEndPos = scimoz.getLineEndPosition(lineNo);
        var text = scimoz.getTextRange(lineStartPos, lineEndPos);
        var cursorPos = currentPos - lineStartPos;
        var wordStartPos = text.search(/\w/);
        var wordEndPos = -1;
        var word = "";
        while (text && (wordStartPos >= 0)) {
            wordEndPos = text.search(/\W/);
            if (wordEndPos < 0) {
                word = text.substr(wordStartPos);
                break;
            }
            word = text.substr(wordStartPos, wordEndPos);
            // XXX - Could be better:
            // This should really be (cursorPos < wordEndPos)
            // But the find system highlights the word and moves the cursor
            // past the end of the word, so under repeating this search, the
            // word is lost and the next word in the line becomes the searched
            // for word.
            if (cursorPos <= wordEndPos) {
                break;
            }
            //dump("Word now: " + word + "\n");
            // Update the variables, moving target here
            text = text.substr(wordEndPos + 1);
            cursorPos -= (word.length + 1);
            wordStartPos = text.search(/\w/);
        }
        if (word.length > 0) {
            //dump("Find word: " + word + "\n");
            // Go and find the next sucker then
            gVimController._searchDirection = searchDirection;
            gVimController.performSearch(scimoz, word);
        } else {
            StatusBar_AddMessage("Search: No word found under the cursor.", 3000, true);
        }
    } catch (e) {
        vimlog.exception(e);
    }
}

function cmd_vim_findWordUnderCursor(scimoz) {
    _cmd_vim_findWordUnderCursor_wrapper(scimoz, VimController.SEARCH_FORWARD);
}

function cmd_vim_findWordUnderCursorBack(scimoz) {
    _cmd_vim_findWordUnderCursor_wrapper(scimoz, VimController.SEARCH_BACKWARD);
}

// Visual mode
function _toggleVisualMode(scimoz, visual_type) {
    if ((gVimController.mode != VimController.MODE_VISUAL) ||
        (gVimController._visualMode != visual_type)) {
        var wasInDifferentVisualMode = (gVimController.mode == VimController.MODE_VISUAL);
        gVimController._visualMode = visual_type;
        gVimController.mode = VimController.MODE_VISUAL;
        if (visual_type == VimController.VISUAL_CHAR) {
            scimoz.selectionMode = scimoz.SC_SEL_STREAM;
        } else if (visual_type == VimController.VISUAL_LINE) {
            scimoz.selectionMode = scimoz.SC_SEL_LINES;
        } else if (visual_type == VimController.VISUAL_BLOCK) {
            scimoz.selectionMode = scimoz.SC_SEL_RECTANGLE;
        }
        if (wasInDifferentVisualMode) {
            // Ensure the status bar shows the correct visual mode
            gVimController.updateStatusBarMessage();
        }
    } else {
        gVimController.mode = VimController.MODE_NORMAL;
    }
}

function cmd_vim_toggleVisualMode(scimoz) {
    _toggleVisualMode(scimoz, VimController.VISUAL_CHAR);
}

function cmd_vim_toggleVisualLineMode(scimoz) {
    _toggleVisualMode(scimoz, VimController.VISUAL_LINE);
}

function cmd_vim_toggleVisualBlockMode(scimoz) {
    _toggleVisualMode(scimoz, VimController.VISUAL_BLOCK);
}

// Special commands
function cmd_vim_replaceChar(scimoz) {
    gVimController.mode = VimController.MODE_REPLACE_CHAR;
}

function cmd_vim_overtype(scimoz, turnOn) {
    try {
        if (typeof(turnOn) == 'undefined' || turnOn) {
            scimoz.overtype = 1;
            gVimController.mode = VimController.MODE_OVERTYPE;
        } else {
            scimoz.overtype = 0;
        }
    } catch (e) {
        vimlog.exception(e);
    }
}

function cmd_vim_dedent(scimoz) {
    try {
        command_doCommand('cmd_home');
        command_doCommand('cmd_dedent');
    } catch (e) {
        vimlog.exception(e);
    }
}

function cmd_vim_indent(scimoz) {
    try {
        command_doCommand('cmd_home');
        command_doCommand('cmd_indent');
    } catch (e) {
        vimlog.exception(e);
    }
}

//
// Special handling for when in the command mode and search mode
//
function cmd_vim_enterCommandMode(scimoz) {
    try {
        gVimController.inputBufferStart(":");
        gVimController.mode = VimController.MODE_COMMAND;
    } catch (e) {
        vimlog.exception(e);
    }
}

function cmd_vim_enterSearchForward(scimoz) {
    gVimController.initSearchForward();
}

function cmd_vim_enterSearchBackward(scimoz) {
    gVimController.initSearchBackward();
}

// Has to be outside the class/object
function vim_InputBuffer_KeyPress(event)
{
    try {
        var stopEvent = false;
        var keyCode = event.keyCode;
        if ((keyCode == event.DOM_VK_ENTER) ||
            (keyCode == event.DOM_VK_RETURN)) {

            var value = gVimController.inputBufferFinish();
            var returnToMode = VimController.MODE_NORMAL;
            if (value.length > 0) {
                value = value.substr(1);
            }
            var scimoz = gViewMgr.currentView.scintilla.scimoz;
            var orig_currentPos = scimoz.currentPos;
            var orig_anchorPos = scimoz.anchor;
            if (gVimController.mode == VimController.MODE_COMMAND) {
                gVimController.findAndRunCommand(value);
            } else if (gVimController.mode == VimController.MODE_SEARCH) {
                // Command mode "/" or "?"
                // We have a search, set up the findSvc and perform the search
                gVimController.performSearch(scimoz, value);
                if (gVimController._lastMode == VimController.MODE_VISUAL) {
                    returnToMode = VimController.MODE_VISUAL;
                }
            }
            // Save into the buffer history
            gVimController.inputBufferSaveHistory(value);

            gVimController.mode = returnToMode;
            gVimController.updateCursorAndSelection(scimoz, orig_currentPos,
                                                    null, orig_anchorPos, true);
            gViewMgr.currentView.setFocus();
            stopEvent = true;
        } else if ((keyCode == event.DOM_VK_ESCAPE) ||
            ((keyCode == event.DOM_VK_BACK_SPACE) &&
             (gVimController.inputBuffer.value.length <= 1))) {
            gVimController.inputBufferFinish();
            gVimController.mode = VimController.MODE_NORMAL;
            stopEvent = true;
            gViewMgr.currentView.setFocus();
        } else if (keyCode == event.DOM_VK_UP) {
            gVimController.inputBufferPrevious();
            stopEvent = true;
        } else if (keyCode == event.DOM_VK_DOWN) {
            gVimController.inputBufferNext();
            stopEvent = true;
        }

        if (stopEvent) {
            event.stopPropagation();
            event.preventDefault();
            event.cancelBubble = true;
        }
    } catch (e) {
        vimlog.exception(e);
    }
}

function vim_InputBuffer_OnFocus(event)
{
    //try {
    //    // if it is not active the hidden input buffer should not have the focus
    //    if (!gVimController._inputBuffer_active && gViewMgr.currentView) {
    //        // This has to be in a timeout for the controllers to work right.
    //        window.setTimeout('gViewMgr.currentView.setFocus();', 1)
    //    }
    //} catch (e) {
    //    vimlog.exception(e);
    //}
}
