/* Copyright (c) 2006 ActiveState Software Inc.
   See the file LICENSE.txt for licensing information. */

/*
 * komodoBaseTreeView - Base tree view written in javascript.
 *
 * Contributers:
 *  - Todd Whiteman (ToddW@ActiveState.com)
 */


/*
 * Sample Usage:

XUL:

    <tree id="sample_tree" flex="1">
        <treecols>
          <treecol id="sample_treecol_name" flex="1" label="Name"/>
          <splitter class="tree-splitter"/>
          <treecol id="sample_treecol_value" flex="1" label="Name"/>
        </treecols>
        <!-- Tree children are dynamically generated, from our own custom tree view -->
        <treechildren/>
    </tree>


JavaScript:

    function SampleTreeView(initial_rows) {
        if (!initial_rows) {
            // Default value then
            this._rows = [];
        } else {
            this._rows = initial_rows;
        }
    }
    // The following two lines ensure proper inheritance (see Flanagan, p. 144).
    SampleTreeView.prototype = new komodoBaseTreeView();
    SampleTreeView.prototype.constructor = SampleTreeView;

    // Override getCellText method for assigning the celltext
    SampleTreeView.prototype.getCellText = function(row, column) {
        // forRow is an item of the _rows array
        var forRow = this._rows[row];
        switch (column.id) {
            case 'sample_treecol_name':
                return forRow[0];
            case 'sample_treecol_value':
                return forRow[1];
        }
        return "(Unknown column: " + column.id + ")";
    };


    // Load the tree view and assign it to a tree element
    var sampleTreeView = new SampleTreeView();
    var sampleTree = document.getElementById("sample_tree");
    sampleTree.treeBoxObject.view = sampleTreeView;
 */

function komodoBaseTreeView(initial_rows) {
    if (!initial_rows) {
        // Default value then
        this._rows = [];
    } else {
        this._rows = initial_rows;
    }
    this._debug = 0;
}

komodoBaseTreeView.SORT_DESCENDING = 0;
komodoBaseTreeView.SORT_ASCENDING = 1;

komodoBaseTreeView.prototype = {
  // The nsITreeView object for the tree view
    /* void QueryInterface(in nsIIDRef uuid,
       [iid_is(uuid),retval] out nsQIResult result); */
    QueryInterface: function(iid) {
        if (!iid.equals(Components.interfaces.nsITreeView) &&
            !iid.equals(Components.interfaces.nsISupports)) {
                throw Components.results.NS_ERROR_NO_INTERFACE;
            }
        return this;
    },

    // Necessary function the child class must override and implement
    getCellText : function(row, column) {
        return "(Child class must implement this)";
    },

    // Everything below here is just to satisfy the interface, override as
    // needed.
    get rowCount() { return this._rows.length; },
    tree : null,
    isSeparator : function(index) {return false;},
    isContainer : function(index) {return false;},
    setTree : function(out) { this.tree = out; },
    getRowProperties : function(row,prop) {},
    getColumnProperties : function(column,prop){},
    getCellProperties : function(row,column,properties) {},
    isContainer : function(row) {return false;},
    cycleCell: function(row, colId) {},
    selectionChanged : function() {},
    performAction : function(action) {},
    isSorted : function() {return false;},
    getImageSrc : function() {return null;},
    cycleHeader : function() {},

    /* Own defined methods */

    /*
     * Function for updating the tree rows, must be called to set what the
     * tree contains in order to then display tree information.
     * 
     * @param rows {Array} The array of items/rows for the tree.
     * @param doReSort {bool} Whether to sort the data after it's set.
     */
    setTreeRows : function(rows, doReSort /* false */) {
        // Clearing the selection
        this.selection.clearSelection();
        // Using rowCountChanged to notify rows were removed
        this.tree.rowCountChanged(0, -this._rows.length);
        this._rows = rows;
        // Using rowCountChanged to notify rows were added
        this.tree.rowCountChanged(0, this._rows.length);
        if (doReSort) {
            this.reSort();
        }
    },

    /* Sorting */
    _sortColumn : null,
    _sortDirection : -1,
    _sortCaseInsensitive : true,
    setSortDirection: function(column, newSortDirection) {
        // Remove old sort direction
        if (this._sortColumn && this._sortColumn != column) {
            this._sortColumn.removeAttribute("sortDirection");
        }

        // Update the sort direction on the sorted column
        if (newSortDirection == komodoBaseTreeView.SORT_ASCENDING) {
            column.setAttribute("sortDirection", "ascending");
        } else {
            column.setAttribute("sortDirection", "descending");
        }
        this._sortColumn = column;
        this._sortDirection = newSortDirection;
    },
    sortByColumn : function(col, sortDirection)
    {
        if (!col) {
            col = this.tree.columns.getFirstColumn().element;
        }
        if (typeof(sortDirection) == 'undefined') {
            var colElem;
            var sortSetting;
            for (var i=0 ; i < this.tree.columns.count; i++) {
                colElem = this.tree.columns.getColumnAt(i).element;
                sortSetting = colElem.getAttribute("sortDirection");
                if (sortSetting == "ascending") {
                    sortDirection = komodoBaseTreeView.SORT_ASCENDING;
                    break;
                } else if (sortSetting == "descending") {
                    sortDirection = komodoBaseTreeView.SORT_DESCENDING;
                    break;
                }
            }
            // No setting already in the column headers, go ascending
            if (typeof(sortDirection) == 'undefined') {
                sortDirection = komodoBaseTreeView.SORT_ASCENDING;
            }
        }

        // Function to compare two row values, compare objects are:
        //  { "data":     Text to compare,
        //    "row":      The original row the text came from
        //    "oldindex": Old selected row index
        //  }
        function _doDataCompare(compare1, compare2)
        {
            if (compare1 && compare2) {
                if (compare1.data > compare2.data) {
                    return -1;
                }
            }
            return 1;
        };

        //dump("Sorting direction: " + this._sortDirection + "\n");
        //dump("Sorting by column: " + col.id + "\n");
        /* if it's the same as last sorted column, just reverse the order */
        if (this._sortColumn == col) {
            //dump("Just reversing sort order for existing column.\n");
            var newCurrentIndex = (this._rows.length - this.selection.currentIndex) - 1;
            this._rows.reverse();
            this.setTreeRows(this._rows, /* doReSort */ false);
            this.selection.currentIndex = newCurrentIndex;
            this.tree.ensureRowIsVisible(newCurrentIndex);
            if (this._sortDirection == komodoBaseTreeView.SORT_ASCENDING) {
                this.setSortDirection(col, komodoBaseTreeView.SORT_DESCENDING);
            } else {
                this.setSortDirection(col, komodoBaseTreeView.SORT_ASCENDING);
            }
            return;
        } else if (typeof(sortDirection) == 'undefined') {
            sortDirection = komodoBaseTreeView.SORT_ASCENDING;
        }

        // Get all the row data we need to compare with
        var unsorted_row_data = [];
        // Convert all fields to a string value
        for (var i=0; i < this._rows.length; i++) {
            if (this._sortCaseInsensitive) {
                unsorted_row_data[i] = { "data": this.getCellText(i, col).toLowerCase(),
                                         "row": this._rows[i],
                                         "oldindex": i };
            } else {
                unsorted_row_data[i] = { "data": this.getCellText(i, col),
                                         "row": this._rows[i],
                                         "oldindex": i };
            }
        }

        // Sort the rows, using our own customized sorting function
        //dump("Beginning sort of " + unsorted_row_values.length + " rows...\n");
        unsorted_row_data.sort(_doDataCompare);
        if (sortDirection == komodoBaseTreeView.SORT_DESCENDING) {
            // Reverse the sorted order
            unsorted_row_data.reverse();
        }

        // Dump sorted rows back into self._rows
        var newCurrentIndex = -1;
        var oldCurrentIndex = this.selection.currentIndex;
        var sorted_rows = [];
        for (var i=0; i < unsorted_row_data.length; i++) {
            sorted_rows.push(unsorted_row_data[i].row);
            // Keep track of the current row selected
            if (oldCurrentIndex == sorted_rows[i].oldindex) {
                newCurrentIndex = i;
            }
        }
        this.setTreeRows(sorted_rows, /* doReSort */ false);

        this.selection.currentIndex = newCurrentIndex;
        this.tree.ensureRowIsVisible(newCurrentIndex);
        this.setSortDirection(col, sortDirection);
    },
    reSort: function() {
        var sortcol = this._sortColumn;
        this._sortColumn = null;
        this.sortByColumn(sortcol);
    },
    sortAscending: function(column) {
        this.sortByColumn(column, komodoBaseTreeView.SORT_ASCENDING);
    },
    sortDescending: function(column) {
        this.sortByColumn(column, komodoBaseTreeView.SORT_DESCENDING);
    }
}
